from .dynamic_biosphere_builder import DynamicBiosphereBuilder
from .edge_extractor import EdgeExtractor
from .helper_classes import SetList
from .matrix_modifier import MatrixModifier
from .timeline_builder import TimelineBuilder
from .timex_lca import TimexLCA

__version__ = "0.3.4"



import math
from datetime import datetime

import bw2data as bd
import pytest
from bw2data.tests import bw2test

from bw_timex import TimexLCA


@pytest.fixture
@bw2test
def wastes_db():
    bd.projects.set_current("__wastes_db__")
    biosphere = bd.Database("biosphere")
    biosphere.write(
        {
            ("biosphere", "CO2"): {
                "type": "emission",
                "name": "carbon dioxide",
            },
        }
    )

    foreground = bd.Database("foreground")
    foreground.register()
    background_2020 = bd.Database("background_2020")
    background_2020.register()

    node_co2 = biosphere.get("CO2")

    # Create nodes
    fu1 = foreground.new_node("fu1", name="fu1", unit="unit")
    fu1["reference product"] = "fu1"
    fu1.save()

    fu2 = foreground.new_node("fu2", name="fu2", unit="unit")
    fu2["reference product"] = "fu2"
    fu2.save()

    fu3 = foreground.new_node("fu3", name="fu3", unit="unit")
    fu3["reference product"] = "fu3"
    fu3.save()

    fu4 = foreground.new_node("fu4", name="fu4", unit="unit")
    fu4["reference product"] = "fu4"
    fu4.save()

    fu5 = foreground.new_node("fu5", name="fu5", unit="unit")
    fu5["reference product"] = "fu5"
    fu5.save()

    some_waste_producing_intermediate_process = foreground.new_node(
        "some_waste_producing_intermediate_process",
        name="some_waste_producing_intermediate_process",
        unit="unit",
    )
    some_waste_producing_intermediate_process["reference product"] = (
        "some_waste_producing_intermediate_process"
    )
    some_waste_producing_intermediate_process.save()

    foreground_waste_treatment_process = foreground.new_node(
        "other_waste_producing_intermediate_process",
        name="other_waste_producing_intermediate_process",
        unit="unit",
    )
    foreground_waste_treatment_process["reference product"] = (
        "other_waste_producing_intermediate_process"
    )
    foreground_waste_treatment_process.save()

    some_normal_intermediate_process = foreground.new_node(
        "some_normal_intermediate_process",
        name="some_normal_intermediate_process",
        unit="unit",
    )
    some_normal_intermediate_process["reference product"] = (
        "some_normal_intermediate_process"
    )
    some_normal_intermediate_process.save()

    used_car = foreground.new_node("used_car", name="used_car", unit="unit")
    used_car["reference product"] = "used_car"
    used_car.save()

    car_treatment_2020 = background_2020.new_node(
        "car_treatment", name="car_treatment", unit="car_treatment"
    )
    car_treatment_2020["reference product"] = "car_treatment"
    car_treatment_2020.save()

    some_other_background_process = background_2020.new_node(
        "some_other_background_process",
        name="some_other_background_process",
        unit="some_other_background_process",
    )
    some_other_background_process["reference product"] = "some_other_background_process"
    some_other_background_process.save()

    # Edges
    fu1.new_edge(input=fu1, amount=1, type="production").save()
    fu1.new_edge(input=used_car, amount=-1, type="technosphere").save()

    fu2.new_edge(input=fu2, amount=1, type="production").save()
    fu2.new_edge(
        input=some_waste_producing_intermediate_process, amount=-1, type="technosphere"
    ).save()

    fu3.new_edge(input=fu3, amount=1, type="production").save()
    fu3.new_edge(input=car_treatment_2020, amount=-1, type="technosphere").save()

    fu4.new_edge(input=fu4, amount=1, type="production").save()
    fu4.new_edge(
        input=some_normal_intermediate_process, amount=-1, type="technosphere"
    ).save()

    fu5.new_edge(input=fu5, amount=1, type="production").save()
    fu5.new_edge(
        input=foreground_waste_treatment_process, amount=1, type="technosphere"
    ).save()

    some_waste_producing_intermediate_process.new_edge(
        input=some_waste_producing_intermediate_process, amount=1, type="production"
    ).save()
    some_waste_producing_intermediate_process.new_edge(
        input=car_treatment_2020, amount=-3, type="technosphere"
    ).save()

    foreground_waste_treatment_process.new_edge(
        input=foreground_waste_treatment_process, amount=-1, type="production"
    ).save()
    foreground_waste_treatment_process.new_edge(
        input=some_waste_producing_intermediate_process, amount=4, type="technosphere"
    ).save()

    some_normal_intermediate_process.new_edge(
        input=some_normal_intermediate_process, amount=1, type="production"
    ).save()
    some_normal_intermediate_process.new_edge(
        input=some_other_background_process, amount=3, type="technosphere"
    ).save()

    used_car.new_edge(input=used_car, amount=-1, type="production").save()
    used_car.new_edge(input=car_treatment_2020, amount=-3, type="technosphere").save()

    car_treatment_2020.new_edge(
        input=car_treatment_2020, amount=-1, type="production"
    ).save()
    car_treatment_2020.new_edge(input=node_co2, amount=1, type="biosphere").save()

    some_other_background_process.new_edge(
        input=some_other_background_process, amount=-1, type="production"
    ).save()
    some_other_background_process.new_edge(
        input=node_co2, amount=1, type="biosphere"
    ).save()

    bd.Method(("GWP", "example")).write(
        [
            (("biosphere", "CO2"), 1),
        ]
    )

    for db in bd.databases:
        bd.Database(db).register()
        bd.Database(db).process()


def test_signs_fu_to_car_to_treatment(wastes_db):
    method = ("GWP", "example")
    database_dates = {
        "background_2020": datetime.strptime("2020", "%Y"),
        "foreground": "dynamic",
    }
    fu = ("foreground", "fu1")
    tlca = TimexLCA({fu: 1}, method, database_dates)
    tlca.build_timeline()
    tlca.lci()
    tlca.static_lcia()

    assert math.isclose(tlca.base_lca.score, tlca.static_score, rel_tol=1e-9)


def test_signs_fu_to_intermediate_to_treatment(wastes_db):
    method = ("GWP", "example")
    database_dates = {
        "background_2020": datetime.strptime("2020", "%Y"),
        "foreground": "dynamic",
    }
    fu = ("foreground", "fu2")
    tlca = TimexLCA({fu: 1}, method, database_dates)
    tlca.build_timeline()
    tlca.lci()
    tlca.static_lcia()

    assert math.isclose(tlca.base_lca.score, tlca.static_score, rel_tol=1e-9)


def test_signs_fu_to_treatment(wastes_db):
    method = ("GWP", "example")
    database_dates = {
        "background_2020": datetime.strptime("2020", "%Y"),
        "foreground": "dynamic",
    }
    fu = ("foreground", "fu3")
    tlca = TimexLCA({fu: 1}, method, database_dates)
    tlca.build_timeline()
    tlca.lci()
    tlca.static_lcia()

    assert math.isclose(tlca.base_lca.score, tlca.static_score, rel_tol=1e-9)


def test_signs_fu_to_normal_intermediate_to_normal_background(wastes_db):
    method = ("GWP", "example")
    database_dates = {
        "background_2020": datetime.strptime("2020", "%Y"),
        "foreground": "dynamic",
    }
    fu = ("foreground", "fu4")
    tlca = TimexLCA({fu: 1}, method, database_dates)
    tlca.build_timeline()
    tlca.lci()
    tlca.static_lcia()

    assert math.isclose(tlca.base_lca.score, tlca.static_score, rel_tol=1e-9)


def test_signs_fu_with_unusual_positives_and_negatives(wastes_db):
    method = ("GWP", "example")
    database_dates = {
        "background_2020": datetime.strptime("2020", "%Y"),
        "foreground": "dynamic",
    }
    fu = ("foreground", "fu5")
    tlca = TimexLCA({fu: 1}, method, database_dates)
    tlca.build_timeline()
    tlca.lci()
    tlca.static_lcia()

    assert math.isclose(tlca.base_lca.score, tlca.static_score, rel_tol=1e-9)



import json
from datetime import datetime, timedelta
from typing import Callable, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from bw2data.backends import ActivityDataset as AD
from bw2data.backends.proxies import Exchange
from bw2data.backends.schema import ExchangeDataset
from bw2data.errors import MultipleResults, UnknownObject
from bw_temporalis import TemporalDistribution, easy_timedelta_distribution
from IPython.display import Javascript, display
from ipywidgets import (
    Button,
    Dropdown,
    FloatSlider,
    HBox,
    IntSlider,
    IntText,
    Label,
    Layout,
    Output,
    Textarea,
    ToggleButtons,
    VBox,
)
from loguru import logger

time_res_mapping_strftime = {
    "year": "%Y",
    "month": "%Y%m",
    "day": "%Y%m%d",
    "hour": "%Y%m%d%H",
}


def extract_date_as_integer(dt_obj: datetime, time_res: Optional[str] = "year") -> int:
    """
    Converts a datetime object to an integer for a given temporal resolution `time_res`

    Parameters
    ----------

    dt_obj : Datetime object.
        Datetime object to be converted to an integer.

    time_res : str, optional
        time resolution to be returned: year=YYYY, month=YYYYMM, day=YYYYMMDD, hour=YYYYMMDDHH

    Returns
    -------
    date_as_integer : int
        Datetime object converted to an integer in the format of time_res

    """
    if time_res not in time_res_mapping_strftime:
        available = list(time_res_mapping_strftime.keys())
        raise ValueError(
            f"Invalid time_res: '{time_res}'. Please choose from: {available}."
        )
    formatted_date = dt_obj.strftime(time_res_mapping_strftime[time_res])
    date_as_integer = int(formatted_date)

    return date_as_integer


def extract_date_as_string(timestamp: datetime, temporal_grouping: str) -> str:
    """
    Extracts the grouping date as a string from a datetime object, based on the chosen temporal
    grouping. E.g. for `temporal_grouping` = 'month', and `timestamp` = 2023-03-29T01:00:00, it
    extracts the string '202303'.


    Parameters
    ----------
    timestamp : Datetime object
        Datetime object to be converted to a string.
    temporal_grouping : str
        Temporal grouping for the date string. Options are: 'year', 'month', 'day', 'hour'


    Returns
    -------
    date_as_string: str
        Date as a string in the format of the chosen temporal grouping.
    """

    if temporal_grouping not in time_res_mapping_strftime.keys():
        raise ValueError(
            f'temporal_grouping: {temporal_grouping} is not a valid option. Please \
            choose from: {list(time_res_mapping_strftime.keys())}, defaulting to "year"',
        )
    return timestamp.strftime(time_res_mapping_strftime[temporal_grouping])


def convert_date_string_to_datetime(temporal_grouping, date_string) -> datetime:
    """
    Converts the string of a date to datetime object.
    e.g. for `temporal_grouping` = 'month', and `date_string` = '202303', it extracts 2023-03-01

    Parameters
    ----------
    temporal_grouping : str
        Temporal grouping for the date string. Options are: 'year', 'month', 'day', 'hour'
    date_string : str
        Date as a string

    Returns
    -------
    datetime
        Datetime object of the date string at the chosen temporal resolution.
    """
    time_res_dict = {
        "year": "%Y",
        "month": "%Y%m",
        "day": "%Y%m%d",
        "hour": "%Y%m%d%H",
    }

    if temporal_grouping not in time_res_dict.keys():
        raise ValueError(
            f'temporal grouping: {temporal_grouping} is not a valid option. Please \
            choose from: {list(time_res_dict.keys())}, defaulting to "year"',
        )
    return datetime.strptime(date_string, time_res_dict[temporal_grouping])


def round_datetime(date: datetime, resolution: str) -> datetime:
    """
    Round a datetime object based on a given resolution

    Parameters
    ----------
    date : datetime
        datetime object to be rounded
    resolution: str
        Temporal resolution to round the datetime object to. Options are: 'year', 'month', 'day' and
        'hour'.

    Returns
    -------
    datetime
        rounded datetime object
    """
    if resolution == "year":
        mid_year = pd.Timestamp(f"{date.year}-07-01")
        return (
            pd.Timestamp(f"{date.year+1}-01-01")
            if date >= mid_year
            else pd.Timestamp(f"{date.year}-01-01")
        )

    if resolution == "month":
        start_of_month = pd.Timestamp(f"{date.year}-{date.month}-01")
        next_month = start_of_month + pd.DateOffset(months=1)
        mid_month = start_of_month + (next_month - start_of_month) / 2
        return next_month if date >= mid_month else start_of_month

    if resolution == "day":
        start_of_day = datetime(date.year, date.month, date.day)
        mid_day = start_of_day + timedelta(hours=12)
        return start_of_day + timedelta(days=1) if date >= mid_day else start_of_day

    if resolution == "hour":
        start_of_hour = datetime(date.year, date.month, date.day, date.hour)
        mid_hour = start_of_hour + timedelta(minutes=30)
        return start_of_hour + timedelta(hours=1) if date >= mid_hour else start_of_hour

    raise ValueError("Resolution must be one of 'year', 'month', 'day', or 'hour'.")


def add_flows_to_characterization_functions(
    flows: Union[str, List[str]],
    func: Callable,
    characterization_functions: Optional[dict] = dict(),
) -> dict:
    """
    Add a new flow or a list of flows to the available characterization functions.

    Parameters
    ----------
    flows : Union[str, List[str]]
        Flow or list of flows to be added to the characterization function dictionary.
    func : Callable
        Dynamic characterization function for flow.
    characterization_functions : dict, optional
        Dictionary of flows and their corresponding characterization functions. Default is an empty
        dictionary.

    Returns
    -------
    dict
        Updated characterization function dictionary with the new flow(s) and function(s).
    """

    # Check if the input is a single flow (str) or a list of flows (List[str])
    if isinstance(flows, str):
        # It's a single flow, add it directly
        characterization_functions[flows] = func
    elif isinstance(flows, list):
        # It's a list of flows, iterate and add each one
        for flow in flows:
            characterization_functions[flow] = func

    return characterization_functions


def resolve_temporalized_node_name(code: str) -> str:
    """
    Getting the name of a node based on the code only.
    Works for non-unique codes if the name is the same across all databases.

    Parameters
    ----------
    code: str
        Code of the node to resolve.

    Returns
    -------
    str
        Name of the node.
    """
    qs = AD.select().where(AD.code == code)
    names = set([obj.name for obj in qs])
    if len(qs) > 1:
        if len(names) > 1:
            raise ValueError(
                "Found multiple names for the given code: {}".format(names)
            )
    elif not qs:
        raise UnknownObject
    return names.pop()


def plot_characterized_inventory_as_waterfall(
    lca_obj,
    static_scores=None,
    prospective_scores=None,
    order_stacked_activities=None,
):
    """
    Plot a stacked waterfall chart of characterized inventory data. As comparison,
    static and prospective scores can be added. Only works for metric GWP at the moment.

    Parameters
    ----------
    lca_obj : TimexLCA
        LCA object with characterized inventory data.
    static_scores : dict, optional
        Dictionary of static scores. Default is None.
    prospective_scores : dict, optional
        Dictionary of prospective scores. Default is None.
    order_stacked_activities : list, optional
        List of activities to order the stacked bars in the waterfall plot. Default is None.

    Returns
    -------
    None
        plots the waterfall chart.

    """
    if not hasattr(lca_obj, "characterized_inventory"):
        raise ValueError("LCA object does not have characterized inventory data.")

    if not hasattr(lca_obj, "activity_time_mapping"):
        raise ValueError("Make sure to pass an instance of a TimexLCA.")

    time_res_dict = {
        "year": "%Y",
        "month": "%Y-%m",
        "day": "%Y-%m-%d",
        "hour": "%Y-%m-%d %H",
    }

    plot_data = lca_obj.characterized_inventory.copy()

    plot_data["year"] = plot_data["date"].dt.strftime(
        time_res_dict[lca_obj.temporal_grouping]
    )  # TODO make temporal resolution flexible

    # Optimized activity label fetching using the TimexLCA's built-in method
    unique_activities = plot_data["activity"].unique()
    activity_labels = {
        idx: lca_obj.get_activity_name_from_time_mapped_id(idx)
        for idx in unique_activities
    }
    plot_data["activity_label"] = plot_data["activity"].map(activity_labels)

    plot_data = plot_data.groupby(["year", "activity_label"], as_index=False)[
        "amount"
    ].sum()
    pivoted_data = plot_data.pivot(
        index="year", columns="activity_label", values="amount"
    )

    combined_data = []
    # Adding exchange_scores as a static column
    if static_scores:
        static_data = pd.DataFrame(
            static_scores.items(), columns=["activity_label", "amount"]
        )
        static_data["year"] = "static"
        pivoted_static_data = static_data.pivot(
            index="year", columns="activity_label", values="amount"
        )
        combined_data.append(pivoted_static_data)

    combined_data.append(pivoted_data)  # making sure the order is correct

    # Adding exchange_scores as a prospective column
    if prospective_scores:
        prospective_data = pd.DataFrame(
            prospective_scores.items(), columns=["activity_label", "amount"]
        )
        prospective_data["year"] = "prospective"
        pivoted_prospective_data = prospective_data.pivot(
            index="year", columns="activity_label", values="amount"
        )
        combined_data.append(pivoted_prospective_data)

    combined_df = pd.concat(combined_data, axis=0)

    if order_stacked_activities:
        combined_df = combined_df[
            order_stacked_activities
        ]  # change order of activities in the stacked bars of the waterfall

    # Calculate the bottom for only the dynamic data
    dynamic_bottom = pivoted_data.sum(axis=1).cumsum().shift(1).fillna(0)

    if static_scores and prospective_scores:
        bottom = pd.concat([pd.Series([0]), dynamic_bottom, pd.Series([0])])
    elif static_scores:
        bottom = pd.concat([pd.Series([0]), dynamic_bottom])
    elif prospective_scores:
        bottom = pd.concat([dynamic_bottom, pd.Series([0])])
    else:
        bottom = dynamic_bottom

    # Plotting
    ax = combined_df.plot(
        kind="bar",
        stacked=True,
        bottom=bottom,
        figsize=(14, 6),
        edgecolor="black",
        linewidth=0.5,
    )
    ax.set_ylabel("GWP [kg CO2-eq]")
    ax.set_xlabel("")
    plt.xticks(rotation=45, ha="right")

    if static_scores:
        ax.axvline(x=0.5, color="black", linestyle="--", lw=1)
    if prospective_scores:
        ax.axvline(x=len(combined_df) - 1.5, color="black", linestyle="--", lw=1)

    handles, labels = ax.get_legend_handles_labels()
    ax.legend(
        handles[::-1],
        labels[::-1],
        loc="center left",
        bbox_to_anchor=(1.02, 0.5),  # x=1.02 moves it outside, y=0.5 centers vertically
        fontsize="small",
    )
    ax.set_axisbelow(True)
    plt.grid(True)
    plt.show()


def get_exchange(**kwargs) -> Exchange:
    """
    Get an exchange from the database.

    Parameters
    ----------
    **kwargs :
        Arguments to specify an exchange.
            - input_node: Input node object
            - input_code: Input node code
            - input_database: Input node database
            - output_node: Output node object
            - output_code: Output node code
            - output_database: Output node database

    Returns
    -------
    Exchange
        The exchange object matching the criteria.

    Raises
    ------
    MultipleResults
        If multiple exchanges match the criteria.
    UnknownObject
        If no exchange matches the criteria.
    """

    # Process input_node if present
    input_node = kwargs.pop("input_node", None)
    if input_node:
        kwargs["input_code"] = input_node["code"]
        kwargs["input_database"] = input_node["database"]

    # Process output_node if present
    output_node = kwargs.pop("output_node", None)
    if output_node:
        kwargs["output_code"] = output_node["code"]
        kwargs["output_database"] = output_node["database"]

    # Map kwargs to database fields
    mapping = {
        "input_code": ExchangeDataset.input_code,
        "input_database": ExchangeDataset.input_database,
        "output_code": ExchangeDataset.output_code,
        "output_database": ExchangeDataset.output_database,
    }

    # Build query filters
    filters = []
    for key, value in kwargs.items():
        field = mapping.get(key)
        if field is not None:
            filters.append(field == value)

    # Execute query with filters
    qs = ExchangeDataset.select().where(*filters)
    candidates = [Exchange(obj) for obj in qs]
    num_candidates = len(candidates)

    if num_candidates > 1:
        raise MultipleResults(
            f"Found {num_candidates} results for the given search. "
            "Please be more specific or double-check your system model for duplicates."
        )
    if num_candidates == 0:
        raise UnknownObject("No exchange found matching the criteria.")

    return candidates[0]


def add_temporal_distribution_to_exchange(
    temporal_distribution: TemporalDistribution, **kwargs
):
    """
    Adds a temporal distribution to an exchange specified by kwargs.

    Parameters
    ----------
    temporal_distribution : TemporalDistribution
        TemporalDistribution to be added to the exchange.
    **kwargs :
        Arguments to specify an exchange.
            - input_node: Input node object
            - input_id: Input node database ID
            - input_code: Input node code
            - input_database: Input node database
            - output_node: Output node object
            - output_id: Output node database ID
            - output_code: Output node code
            - output_database: Output node database

    Returns
    -------
    None
        The exchange is saved with the temporal distribution.
    """
    exchange = get_exchange(**kwargs)
    exchange["temporal_distribution"] = temporal_distribution
    exchange.save()
    logger.info(f"Added temporal distribution to exchange {exchange}.")


def interactive_td_widget():
    """
    Create an interactive ipywidget for drafting and visualizing temporal distributions and copying 
    them to the clipboard.

    For use in jupyter notebooks.

    Returns
    -------
    ipywidgets.VBox
        Interactive widget for drafting temporal distributions.
    """
    # ---------- Controls ----------
    mode = ToggleButtons(
        options=["Generator", "Manual"], value="Generator", description="Mode"
    )

    # Generator controls
    start = IntText(value=0, description="start")
    end = IntText(value=10, description="end")
    resolution = Dropdown(
        options=[("Years", "Y"), ("Months", "M"), ("Days", "D")],
        value="Y",
        description="resolution",
    )
    steps = IntSlider(
        value=11, min=2, max=11, step=1, description="steps", continuous_update=False
    )
    kind = ToggleButtons(
        options=["uniform", "triangular", "normal"], value="uniform", description="kind"
    )
    # Give wide initial bounds; we'll override on kind changes
    param = FloatSlider(
        value=1.0,
        min=0.01,
        max=50.0,
        step=0.01,
        description="param",
        disabled=True,
        continuous_update=False,
    )

    # Manual controls
    manual_unit = Dropdown(
        options=[
            ("Years", "Y"),
            ("Months", "M"),
            ("Days", "D"),
            ("Hours", "h"),
            ("Minutes", "m"),
            ("Seconds", "s"),
        ],
        value="Y",
        description="resolution",
    )
    dates_text = Textarea(
        value="0, 2, 4, 6, 8, 10",
        description="dates",
        layout=Layout(width="100%", min_height="70px"),
    )
    amounts_text = Textarea(
        value="0.1, 0.1, 0.2, 0.2, 0.2, 0.2",
        description="amounts",
        layout=Layout(width="100%", min_height="70px"),
    )

    for widget in (start, end, resolution, steps, param, manual_unit):
        widget.style.description_width = "initial"

    steps.layout = Layout(width="220px")
    param.layout = Layout(width="220px")
    start.layout = Layout(width="160px")
    end.layout = Layout(width="160px")
    resolution.layout = Layout(width="180px")
    manual_unit.layout = Layout(width="220px")

    copy_btn = Button(description="Copy TD code", button_style="success")
    copy_import_btn = Button(description="Copy TD + imports", button_style="")
    copy_btn.layout = Layout(width="160px")
    copy_import_btn.layout = Layout(width="200px")
    status = Label(value="")
    status.layout = Layout(margin="0 0 0 8px")
    plot_out = Output(layout=Layout(width="100%"))

    # ---------- Helpers ----------
    def _parse_num_list(txt: str, label: str) -> List[float]:
        parts = [p for p in txt.replace(",", " ").split() if p]
        if not parts:
            raise ValueError(f"{label} cannot be empty.")

        values = []
        for p in parts:
            try:
                if ("." in p) or ("e" in p.lower()):
                    values.append(float(p))
                else:
                    values.append(int(p))
            except ValueError as exc:
                raise ValueError(f"Could not parse '{p}' in {label}.") from exc
        return values

    def _format_number(value: float) -> str:
        as_float = float(value)
        if np.isfinite(as_float) and as_float.is_integer():
            return str(int(as_float))
        return (f"{as_float:.6f}").rstrip("0").rstrip(".")

    def _make_td_generator():
        return easy_timedelta_distribution(
            start=min(start.value, end.value),
            end=max(start.value, end.value),
            resolution=resolution.value,
            steps=int(steps.value),
            kind=kind.value,
            param=None if param.disabled else float(param.value),
        )

    def _make_td_manual():
        d = _parse_num_list(dates_text.value, "dates")
        a = _parse_num_list(amounts_text.value, "amounts")
        if len(d) != len(a):
            raise ValueError("dates and amounts must have the same length.")
        if not d:
            raise ValueError("Provide at least one date and amount.")
        date = np.array(d, dtype=f"timedelta64[{manual_unit.value}]")
        amount = np.array(a, dtype=float)
        if np.any(np.isnan(amount)):
            raise ValueError("Amounts must be numeric values.")
        return TemporalDistribution(date=date, amount=amount)

    def _current_td():
        return _make_td_generator() if mode.value == "Generator" else _make_td_manual()

    def _current_resolution_for_graph():
        return resolution.value if mode.value == "Generator" else manual_unit.value

    def _draw_graph(td: TemporalDistribution):
        with plot_out:
            plot_out.clear_output(wait=True)
            fig = plt.figure(figsize=(7, 3))
            td.graph(style="default", resolution=_current_resolution_for_graph())
            plt.show()
            plt.close(fig)
        status.value = (
            f"OK · steps={len(td.amount)} · sum(amount)={float(np.sum(td.amount)):.6f}"
        )

    def refresh_preview(*_):
        try:
            td = _current_td()
            _draw_graph(td)
        except Exception as exc:
            with plot_out:
                plot_out.clear_output(wait=True)
            status.value = f"Error: {exc}"

    # --- robust param updater (avoid value snapping back to 1.0) ---
    def _with_param_unobserved(fn):
        try:
            param.unobserve(refresh_preview, names="value")
            fn()
        finally:
            param.observe(refresh_preview, names="value")

    def _reset_param_for_kind():
        def _apply():
            s, e = sorted([start.value, end.value])
            span = abs(e - s)

            def _set_slider_value(target: float) -> None:
                bounded = min(max(target, param.min), param.max)
                step = param.step or 0
                if step <= 0:
                    param.value = bounded
                    return
                base = param.min
                ticks = round((bounded - base) / step)
                param.value = base + ticks * step

            if kind.value == "uniform":
                param.description = "param"
                param.disabled = True
                param.min = 0.1
                param.max = 50.0
                param.step = 0.1
                param.value = 1.0
                param.layout.display = "none"
                return

            if kind.value == "triangular":
                param.description = "mode"
                if s == e:
                    param.disabled = True
                    exact_value = float(s)
                    param.min = exact_value
                    param.max = exact_value
                    param.step = 1.0
                    param.value = exact_value
                else:
                    param.disabled = False
                    param.min = float(s)
                    param.max = float(e)
                    param.step = max((param.max - param.min) / 20.0, 0.01)
                    _set_slider_value((param.min + param.max) / 2.0)
                param.layout.display = ""
                return

            # normal
            param.description = "std dev"
            param.disabled = False
            span = max(span, 1)
            param.min = 0.02
            param.max = 1.0
            param.step = 0.01
            _set_slider_value(0.15)
            param.layout.display = ""

        _with_param_unobserved(_apply)

    def _code_generator():
        s, e = sorted([start.value, end.value])
        k = kind.value
        p = None if param.disabled else float(param.value)
        code = (
            "td = easy_timedelta_distribution(\n"
            f"    start={s},\n"
            f"    end={e},\n"
            f"    resolution='{resolution.value}',\n"
            f"    steps={int(steps.value)},\n"
            f"    kind='{k}'"
        )
        if p is not None:
            code += f",\n    param={_format_number(p)}"
        code += "\n)"
        return code

    def _code_manual():
        d = _parse_num_list(dates_text.value, "dates")
        a = _parse_num_list(amounts_text.value, "amounts")
        unit = manual_unit.value
        d_str = ", ".join(str(int(x)) for x in d)
        a_str = ", ".join(_format_number(x) for x in a)
        return (
            f"date = np.array([{d_str}], dtype='timedelta64[{unit}]')\n"
            f"amount = np.array([{a_str}], dtype=float)\n"
            "td = TemporalDistribution(date=date, amount=amount)"
        )

    def _build_code(include_imports: bool = False) -> str:
        body = _code_generator() if mode.value == "Generator" else _code_manual()
        if not include_imports:
            return body

        if mode.value == "Generator":
            imports = [
                "from bw_temporalis import easy_timedelta_distribution",
            ]
        else:
            imports = [
                "import numpy as np",
                "from bw_temporalis import TemporalDistribution",
            ]

        return "\n".join(imports + ["", body])

    def _copy_code(include_imports: bool) -> None:
        try:
            code = _build_code(include_imports=include_imports)
            display(Javascript(f"navigator.clipboard.writeText({json.dumps(code)})"))
            suffix = " + imports" if include_imports else ""
            status.value = f"✅ Code{suffix} copied to clipboard!"
        except Exception as exc:
            status.value = f"Error: {exc}"

    # ---------- Updates ----------
    def _on_kind_change(_):
        _reset_param_for_kind()
        refresh_preview()

    def _on_start_end_change(_):
        _reset_param_for_kind()
        refresh_preview()

    for w in (start, end):
        w.observe(_on_start_end_change, names="value")
    kind.observe(_on_kind_change, names="value")
    for w in (resolution, steps):
        w.observe(refresh_preview, names="value")
    param.observe(
        refresh_preview, names="value"
    )  # reattached in _with_param_unobserved
    for w in (mode, manual_unit, dates_text, amounts_text):
        w.observe(refresh_preview, names="value")
    copy_btn.on_click(lambda _: _copy_code(include_imports=False))
    copy_import_btn.on_click(lambda _: _copy_code(include_imports=True))

    # Initial state
    _reset_param_for_kind()
    refresh_preview()

    # ---------- UI ----------
    gen_box = VBox(
        [
            HBox([start, end, resolution], layout=Layout(gap="10px")),
            HBox([steps, param], layout=Layout(gap="10px")),
        ],
        layout=Layout(gap="10px"),
    )
    man_box = VBox(
        [manual_unit, dates_text, amounts_text],
        layout=Layout(gap="8px", width="100%"),
    )

    # Keep steps.max synced to end for nicer defaults
    def _sync_steps_max(_=None):
        new_max = abs(end.value - start.value) + 1
        if steps.max != new_max:
            steps.max = new_max

    _sync_steps_max()
    end.observe(_sync_steps_max, names="value")
    start.observe(_sync_steps_max, names="value")

    buttons_box = HBox(
        [copy_btn, copy_import_btn, status],
        layout=Layout(align_items="center", gap="10px"),
    )

    def _layout_children():
        if mode.value == "Generator":
            return [
                mode,
                kind,
                gen_box,
                buttons_box,
                plot_out,
            ]
        return [
            mode,
            man_box,
            buttons_box,
            plot_out,
        ]

    container = VBox(_layout_children(), layout=Layout(gap="12px", width="100%"))

    def _mode_refresh(_):
        container.children = _layout_children()
        if mode.value == "Generator":
            _sync_steps_max()
            _reset_param_for_kind()
        refresh_preview()

    mode.observe(_mode_refresh, names="value")

    return container



from datetime import datetime
from functools import partial
from itertools import chain
from typing import Callable, Optional

import bw2data as bd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sb
from bw2calc import LCA
from bw2data import (
    Database,
    Method,
    Normalization,
    Weighting,
    databases,
    methods,
    normalizations,
    projects,
    weightings,
)
from bw2data.backends.schema import ActivityDataset as AD
from bw2data.backends.schema import get_id
from bw2data.errors import Brightway2Project
from dynamic_characterization import characterize
from loguru import logger
from peewee import fn
from scipy import sparse

from .dynamic_biosphere_builder import DynamicBiosphereBuilder
from .helper_classes import InterDatabaseMapping, TimeMappingDict
from .matrix_modifier import MatrixModifier
from .timeline_builder import TimelineBuilder
from .utils import (
    convert_date_string_to_datetime,
    extract_date_as_integer,
    round_datetime,
)


class TimexLCA:
    """
    Class to perform time-explicit LCA calculations.

    A TimexLCA contains the LCI of processes occurring at explicit points in time. It tracks the timing of processes,
    relinks their technosphere and biosphere exchanges to match the technology landscape at that point in time,
    and also keeps track of the timing of the resulting emissions. As such, it combines prospective and dynamic LCA
    approaches.

    TimexLCA first calculates a static LCA, which informs a priority-first graph traversal. From the
    graph traversal, temporal relationships between exchanges and processes are derived. Based on
    the timing of the processes, bw_timex matches the processes at the intersection between
    foreground and background to the best available background databases. This temporal relinking is
    achieved by using datapackages to add new time-specific processes. The new processes and their
    exchanges to other technosphere processes or biosphere flows extent the technosphere and
    biosphere matrices.

    Temporal information of both processes and biosphere flows is retained, allowing for dynamic
    LCIA.

    TimexLCA calculates:
     1) a static "base" LCA score (`TimexLCA.base_score`, same as `bw2calc.lca.score`),
     2) a static time-explicit LCA score (`TimexLCA.static_score`), which links LCIs to the
        respective background databases, but without dynamic characterization of the time-explicit inventory
     3) a dynamic time-explicit LCA score (`TimexLCA.dynamic_score`), with dynamic inventory and
        dynamic characterization. These are provided for radiative forcing and GWP but can also be
        user-defined.


    Example
    -------
    >>> demand = {('my_foreground_database', 'my_process'): 1}
    >>> method = ("some_method_family", "some_category", "some_method")
    >>> database_dates = {
            'my_background_database_one': datetime.strptime("2020", "%Y"),
            'my_background_database_two': datetime.strptime("2030", "%Y"),
            'my_background_database_three': datetime.strptime("2040", "%Y"),
            'my_foreground_database':'dynamic'
        }
    >>> tlca = TimexLCA(demand, method, database_dates)
    >>> tlca.build_timeline() # has many optional arguments
    >>> tlca.lci()
    >>> tlca.static_lcia()
    >>> print(tlca.static_score)
    >>> tlca.dynamic_lcia(metric="radiative_forcing") # also available: "GWP"
    >>> print(tlca.dynamic_score)

    """ """"""

    def __init__(
        self,
        demand: dict,
        method: tuple,
        database_dates: dict = None,
    ) -> None:
        """
        Instantiating a `TimexLCA` object calculates a static LCA, initializes time mappings
        for activities and biosphere flows, and stores useful subsets of ids in the
        node_collections.

        Parameters
        ----------
        demand : dict[object: float]
                The demand for which the LCA will be calculated. The keys can be Brightway `Node`
                instances, `(database, code)` tuples, or integer ids.
        method : tuple
                Tuple defining the LCIA method, such as `('foo', 'bar')` or default methods, such as
                `("EF v3.1", "climate change", "global warming potential (GWP100)")`
        database_dates : dict, optional
                Dictionary mapping database names to dates.
        """

        logger.info("Initializing TimexLCA object...")

        self.demand = demand
        self.method = method
        self.database_dates = database_dates

        if not self.database_dates:
            logger.info(
                "No database_dates provided. Treating the databases containing the functional \
                unit as dynamic. No remapping of inventories to time explicit databases will be done."
            )
            self.database_dates = {key[0]: "dynamic" for key in demand.keys()}

        self.check_format_database_dates()

        # Create static_only dict that excludes dynamic processes that will be exploded later.
        # This way we only have the "background databases" that we can later link to from the dates
        # of the timeline.
        self.database_dates_static = {
            k: v for k, v in self.database_dates.items() if isinstance(v, datetime)
        }

        logger.info("Collecting node infos...")
        # Create some collections of nodes that will be useful down the line, e.g. all nodes from
        # the background databases that link to foreground nodes.
        self.create_node_collections()

        self.interdatabase_activity_mapping = InterDatabaseMapping()

        # Getting all nodes from the databases for faster lookup later
        all_nodes = bd.backends.ActivityDataset.select().where(
            bd.backends.ActivityDataset.database << list(self.database_dates.keys())
        )
        self.nodes = {node.id: bd.backends.Activity(node) for node in all_nodes}

        # Build a cache mapping activity code to name for efficient lookups
        # This avoids repeated database queries in plotting and labeling functions
        self._activity_code_to_name_cache = {
            node["code"]: node["name"] for node in self.nodes.values()
        }

    ########################################
    # Main functions to be called by users #
    ########################################

    def build_timeline(
        self,
        starting_datetime: datetime | str = "now",
        temporal_grouping: str = "year",
        interpolation_type: str = "linear",
        edge_filter_function: Callable = None,
        cutoff: float = 1e-9,
        max_calc: int = 2000,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Creates a `TimelineBuilder` instance that does the graph traversal (similar to
        bw_temporalis) and extracts all edges with their temporal information. Creates the
        `TimexLCA.timeline` of technosphere exchanges.

        Parameters
        ----------
        starting_datetime: datetime | str, optional
            Point in time when the demand occurs. This is the initial starting point of the
            graph traversal and the timeline. Something like `"now"` or `"2023-01-01"`.
            Default is `"now"`.
        temporal_grouping : str, optional
            Time resolution for grouping exchanges over time in the timeline. Default is 'year',
            other options are 'month', 'day', 'hour'.
        interpolation_type : str, optional
            Type of interpolation when sourcing the new producers in the time-mapped background
            databases. Default is 'linear', which means linear interpolation between the closest 2
            databases, other options are 'closest', which selects only the closest database.
        edge_filter_function : Callable, optional
            Function to skip edges in the graph traversal. Default is to skip all edges within
            background databases.
        cutoff: float, optional
            The cutoff value for the graph traversal. Default is 1e-9.
        max_calc: float, optional
            The maximum number of calculations to be performed by the graph traversal. Default is
            2000.
        *args : iterable
            Positional arguments for the graph traversal, for `bw_temporalis.TemporalisLCA` passed
            to the `EdgeExtractor` class, which inherits from `TemporalisLCA`. See `bw_temporalis`
            documentation for more information.
        **kwargs : dict
            Additional keyword arguments for the graph traversal, for `bw_temporalis.TemporalisLCA`
            passed to the EdgeExtractor class, which inherits from TemporalisLCA. See bw_temporalis
            documentation for more information.

        Returns
        -------
        pandas.DataFrame:
            A DataFrame containing the timeline of technosphere exchanges

        See also
        --------
        bw_timex.timeline_builder.TimelineBuilder: Class that builds the timeline.

        """
        if edge_filter_function is None:
            logger.info(
                "No edge filter function provided. Skipping all edges in background databases."
            )
            skippable = []
            for db in self.database_dates_static.keys():
                skippable.extend([node.id for node in bd.Database(db)])
            self.edge_filter_function = lambda x: x in skippable
        else:
            self.edge_filter_function = edge_filter_function

        self.starting_datetime = starting_datetime
        self.temporal_grouping = temporal_grouping
        self.interpolation_type = interpolation_type
        self.cutoff = cutoff
        self.max_calc = max_calc

        logger.info("Calculating base LCA...")
        # Calculate static LCA results using a custom prepare_lca_inputs function that includes all
        # background databases in the LCA. We need all the IDs for the time mapping dict.
        fu, data_objs, remapping = self.prepare_base_lca_inputs(
            demand=self.demand, method=self.method
        )
        self.base_lca = LCA(fu, data_objs=data_objs, remapping_dicts=remapping)
        self.base_lca.lci()
        self.base_lca.lcia()

        logger.info("Creating activity time mapping...")
        # Create a time mapping dict that maps each activity to a activity_time_mapping_id in the
        # format (('database', 'code'), datetime_as_integer): time_mapping_id)
        self.activity_time_mapping = TimeMappingDict(
            start_id=bd.backends.ActivityDataset.select(fn.MAX(AD.id)).scalar() + 1
        )  # making sure we get unique ids by counting up from the highest current activity id

        # pre-populate the activity time mapping dict with the static activities.
        # Doing this here because we need the temporal grouping for consistent time resolution.
        self.add_static_activities_to_activity_time_mapping()

        # Create timeline builder that does the graph traversal (similar to bw_temporalis) and
        # extracts all edges with their temporal information. Can later be used to build a timeline
        # with the TimelineBuilder.build_timeline() method.
        self.timeline_builder = TimelineBuilder(
            self.base_lca,
            self.starting_datetime,
            self.edge_filter_function,
            self.database_dates,
            self.database_dates_static,
            self.activity_time_mapping,
            self.node_collections,
            self.nodes,
            self.temporal_grouping,
            self.interpolation_type,
            self.cutoff,
            self.max_calc,
            *args,
            **kwargs,
        )

        self.timeline = self.timeline_builder.build_timeline()

        self.add_interdatabase_activity_mapping_from_timeline()

        return self.timeline[
            [
                "date_producer",
                "producer_name",
                "date_consumer",
                "consumer_name",
                "amount",
                "temporal_market_shares",
            ]
        ]

    def lci(
        self,
        build_dynamic_biosphere: Optional[bool] = True,
        expand_technosphere: Optional[bool] = True,
    ) -> None:
        """
        Calculates the time-explicit LCI.

        There are two ways to generate time-explicit LCIs:
        If `expand_technosphere' is True, the biosphere and technosphere matrices are expanded by inserting
        time-specific processes via the `MatrixModifier` class by calling `TimexLCA.build_datapackage().
        Otherwise ('expand_technosphere' is False), it generates a dynamic inventory directly from the
        timeline without technosphere matrix calculations.

        Next to the choice above concerning how to retrieve the time-explicit inventory, users
        can also decide if they want to retain all temporal information at the biosphere level
        (build_dynamic_biosphere = True).
        Set `build_dynamic_biosphere` to False if you only want to get a new overall score of
        the time-explicit inventory and don't care about the timing of the emissions.
        This saves time and memory.

        Parameters
        ----------
        build_dynamic_biosphere: bool
            if True, build the dynamic biosphere matrix and calculate the dynamic LCI.
            Default is True.
        expand_technosphere: bool
            if True, creates an expanded time-explicit technosphere and biosphere matrix and
            calculates the LCI from it.
            if False, creates no new technosphere, but calculates the dynamic inventory directly
            from the timeline. Building from the timeline currently only works if
            `build_dynamic_biosphere` is also True.

        Returns
        -------
        None
            calls LCI calculations from bw2calc and calculates the dynamic inventory, if
            `build_dynamic_biosphere` is True.

        See also
        --------
        build_datapackage:
            Method to create the datapackages that contain the modifications
            to the technosphere and biosphere matrix using the `MatrixModifier` class.
        calculate_dynamic_inventory:
            Method to calculate the dynamic inventory if `build_dynamic_biosphere` is True.
        """

        if hasattr(self, "dynamic_inventory"):
            del self.dynamic_inventory

        if not expand_technosphere and not build_dynamic_biosphere:
            raise ValueError(
                "Currently, it is not possible to skip the construction of the dynamic \
                biosphere when building the inventories from the timeline.\
                Please either set build_dynamic_biosphere=True or expand_technosphere=True"
            )

        if not hasattr(self, "timeline"):
            raise AttributeError(
                "Timeline not yet built. Call TimexLCA.build_timeline() first."
            )

        # mapping of the demand id to demand time
        self.demand_timing = self.create_demand_timing()

        self.fu, self.data_objs, self.remapping = self.prepare_bw_timex_inputs(
            demand=self.demand,
            method=self.method,
        )

        if expand_technosphere:
            logger.info("Expanding matrices...")
            self.datapackage = self.build_datapackage()
            data_obs = self.data_objs + self.datapackage
            self.expanded_technosphere = True  # set flag for later static lcia usage
        else:  # setup for timeline approach
            logger.warning(
                "Building the dynamic inventory directly from the timeline. This feature is under development.\
                Use at your own risk... and check your results! Disaggregated lci is not yet implemented."
            )
            self.collect_temporalized_processes_from_timeline()
            data_obs = self.data_objs
            self.expanded_technosphere = False  # set flag for later lcia usage

        self.lca = LCA(
            self.fu,
            data_objs=data_obs,
            remapping_dicts=self.remapping,
        )

        logger.info("Calculating dynamic inventory...")
        if not build_dynamic_biosphere:
            self.lca.lci()
        else:  # building dynamic biosphere
            if expand_technosphere:
                self.lca.lci(factorize=True)
                self.calculate_dynamic_inventory(expand_technosphere=True)
                # to get back the original LCI - necessary because we do some redo_lci's in the
                # dynamic inventory calculation
                self.lca.redo_lci(self.fu)
            else:
                self.calculate_dynamic_inventory(expand_technosphere=False)

    def disaggregate_background_lci(self) -> None:
        """
        This method disaggregates the background LCI's of the temporal markets.
        The disaggregated background LCI's allow a contribution analysis on the
        orginal inventory level as compared to the aggregated temporal market emissions.

        Parameters
        ----------
        None

        Returns
        -------
        None
            Stores the disaggregated background inventory in the attribute
            `dynamic_inventory_disaggregated` as a matrix and in `dynamic_inventory_disaggregated_df`
            as a DataFrame.
        """

        if not hasattr(self, "dynamic_inventory"):
            raise AttributeError(
                "Dynamic lci not yet calculated. Call TimexLCA.lci(build_dynamic_biosphere=True) first."
            )
        if not self.expanded_technosphere:
            raise NotImplementedError(
                "Currently the disaggregation of background processes is only possible\n\
                    if the expanded matrix has been built. Please call TimexLCA.lci(expand_technosphere=True) first."
            )
        # create array_dict for fast lookup
        # (key becomes index, value becomes value of 1D array)
        bio_dict_array = np.zeros(
            max(self.lca.dicts.biosphere.reversed.keys()) + 1, dtype=int
        )
        for key, value in self.lca.dicts.biosphere.reversed.items():
            bio_dict_array[key] = value

        # create biosphere_time_mapping_int for fast lookup
        biosphere_time_mapping_int = {
            (key[0], key[1].astype("int64")): value
            for key, value in self.biosphere_time_mapping.items()
        }

        self.dynamic_inventory_disaggregated = self.dynamic_inventory.tocsc()
        # 1) set all temporal market emissions to zero
        for col in self.dynamic_biosphere_builder.temporal_market_cols:
            self.dynamic_inventory_disaggregated.data[
                self.dynamic_inventory_disaggregated.indptr[
                    col
                ] : self.dynamic_inventory_disaggregated.indptr[col + 1]
            ] = 0
        self.dynamic_inventory_disaggregated.eliminate_zeros()
        # 2) add all background inventory to the dynamic inventory for all temporal markets

        self.dynamic_inventory_disaggregated = (
            self.dynamic_inventory_disaggregated.tocoo()
        )

        dynamic_inv_row_ids = self.dynamic_inventory_disaggregated.row.tolist()
        dynamic_inv_col_ids = self.dynamic_inventory_disaggregated.col.tolist()
        dynamic_inv_data = self.dynamic_inventory_disaggregated.data.tolist()

        for id_, lci in self.temporal_market_lcis.items():

            ((_, _), time) = self.activity_time_mapping.reversed[
                id_
            ]  # time of temporal market
            time_in_datetime = convert_date_string_to_datetime(
                self.temporal_grouping, str(time)
            )
            time_int = (
                np.datetime64(time_in_datetime).astype("datetime64[s]").astype("int64")
            )  # now time is a int64 in secs

            lci = lci.tocoo()

            # create list of tuples for fast lookup
            time_array = np.ones(len(lci.row), dtype="int64") * time_int
            list_of_tuples = list(zip(bio_dict_array[lci.row], time_array))

            new_rows = [biosphere_time_mapping_int[x] for x in list_of_tuples]
            dynamic_inv_row_ids.extend(new_rows)
            dynamic_inv_col_ids.extend(lci.col)
            dynamic_inv_data.extend(lci.data)

        # construct the new dynamic inventory including background inventory instead of aggregated temporal market emissions
        dynamic_inventory_disaggregated = sparse.coo_matrix(
            (dynamic_inv_data, (dynamic_inv_row_ids, dynamic_inv_col_ids)),
            shape=self.dynamic_inventory_disaggregated.shape,
        )
        self.dynamic_inventory_disaggregated = dynamic_inventory_disaggregated.tocsr()
        self.dynamic_inventory_disaggregated_df = (
            self.create_dynamic_inventory_dataframe(use_disaggregated_lci=True)
        )

    def static_lcia(self) -> None:
        """
        Calculates static LCIA using time-explicit LCIs with the standard static characterization
        factors of the selected LCIA method using `bw2calc.lcia()`.

        Parameters
        ----------
        None

        Returns
        -------
        None
            Stores the static score in the attribute `static_score`.
        """
        if not hasattr(self, "lca"):
            raise AttributeError("LCI not yet calculated. Call TimexLCA.lci() first.")
        if not self.expanded_technosphere:
            raise ValueError(
                "Currently the static lcia score can only be calculated if the expanded matrix has \
                    been built. Please call TimexLCA.lci(expand_technosphere=True) first."
            )
        self.lca.lcia()

    def dynamic_lcia(
        self,
        metric: str = "radiative_forcing",
        time_horizon: int = 100,
        fixed_time_horizon: bool = False,
        time_horizon_start: datetime = None,
        characterization_functions: dict = None,
        characterization_function_co2: dict = None,
        use_disaggregated_lci: bool = False,
    ) -> pd.DataFrame:
        """
        Calculates dynamic LCIA with the `DynamicCharacterization` class using the dynamic inventory
        and dynamic characterization functions. Dynamic characterization is handled by the separate
        package `dynamic_characterization` (https://dynamic-characterization.readthedocs.io).

        Dynamic characterization functions in the form of a dictionary {biosphere_flow_database_id:
        characterization_function} can be given by the user.
        If none are given, a set of default dynamic characterization functions based on IPCC AR6 are
        provided from `dynamic_characterization` package. These are mapped to the biosphere3 flows
        of the chosen static climate change impact category. If there is no characterization
        function for a biosphere flow, it will be ignored.

        Two dynamic climate change metrics are supported: "GWP" and "radiative_forcing".
        The time horizon for the impact assessment can be set with the `time_horizon` parameter,
        defaulting to 100 years. The `fixed_time_horizon` parameter determines whether the emission
        time horizon for all emissions is calculated from a specific starting point `time_horizon_start`
        (`fixed_time_horizon=True`) or from the time of the emission (`fixed_time_horizon=False`).
        The former is the implementation of the Levasseur approach
        (see https://doi.org/10.1021/es9030003), while the latter is how conventional LCA is done.

        Parameters
        ----------
        metric : str, optional
            the metric for which the dynamic LCIA should be calculated. Default is
            "radiative_forcing". Available: "GWP" and "radiative_forcing"
        time_horizon: int, optional
            the time horizon for the impact assessment. Unit is years. Default is 100.
        fixed_time_horizon: bool, optional
            Whether the emission time horizon for all emissions is calculated from the functional
            unit (fixed_time_horizon=True) or from the time of the emission
            (fixed_time_horizon=False). Default is False.
        time_horizon_start: pd.Timestamp, optional
            The starting timestamp of the time horizon for the dynamic characterization. Only needed
            for fixed time horizons. Default is datetime.now().
        characterization_functions: dict, optional
            Dict of the form {biosphere_flow_database_id: characterization_function}. Default is
            None, which triggers the use of the provided dynamic characterization functions based on
            IPCC AR6 Chapter 7.
        characterization_function_co2: Callable, optional
            Characterization function for CO2 emissions. Necessary if GWP metric is chosen. Default
            is None, which triggers the use of the provided dynamic characterization function of CO2
            based on IPCC AR6 Chapter 7.
        use_disaggregated_lci: bool, optional
            Whether to use the disaggregated background LCI for the dynamic LCIA. Default is False.
            Use True if you want to perform a contribution analysis on the disaggregated background.

        Returns
        -------
        pandas.DataFrame
            A DataFrame with the characterized inventory for the chosen metric and parameters.

        See also
        --------
        dynamic_characterization: Package handling the dynamic characterization: https://dynamic-characterization.readthedocs.io/en/latest/
        """

        if not hasattr(self, "dynamic_inventory"):
            raise AttributeError(
                "Dynamic lci not yet calculated. Call TimexLCA.lci(build_dynamic_biosphere=True) first."
            )

        self.current_metric = metric
        self.current_time_horizon = time_horizon

        if use_disaggregated_lci:
            if not self.expanded_technosphere:
                raise NotImplementedError(
                    "Currently the disaggregation of background processes is only possible if the \
                        expanded matrix has been built. Please call TimexLCA.lci(expand_technosphere=True) first."
                )
            # Check if disaggregated inventory is available
            # otherwise disaggregate the background LCI
            if not hasattr(self, "dynamic_inventory_disaggregated"):
                logger.info("Disaggregating background LCI...")
                self.disaggregate_background_lci()
                logger.info("Background LCI's disaggregated.")
            dynamic_inventory_df = self.dynamic_inventory_disaggregated_df

        else:
            dynamic_inventory_df = self.dynamic_inventory_df

        # Set a default for inventory_in_time_horizon using the full dynamic_inventory_df
        inventory_in_time_horizon = dynamic_inventory_df

        # Round dates to nearest year and sum up emissions for each year
        inventory_in_time_horizon.date = inventory_in_time_horizon.date.apply(
            partial(round_datetime, resolution="year")
        )
        inventory_in_time_horizon = (
            inventory_in_time_horizon.groupby(
                inventory_in_time_horizon.columns.tolist()
            )
            .sum()
            .reset_index()
        )

        # Calculate the latest considered impact date
        t0_date = pd.Timestamp(self.timeline_builder.edge_extractor.t0.date[0])
        latest_considered_impact = t0_date + pd.DateOffset(years=time_horizon)

        # Update inventory_in_time_horizon if a fixed time horizon is used
        if fixed_time_horizon:
            last_emission = dynamic_inventory_df.date.max()
            if latest_considered_impact < last_emission:
                logger.warning(
                    "An emission occurs outside of the specified time horizon and will not be \
                        characterized. Please make sure this is intended."
                )
                inventory_in_time_horizon = dynamic_inventory_df[
                    dynamic_inventory_df.date <= latest_considered_impact
                ]

        if not time_horizon_start:
            time_horizon_start = t0_date

        self.characterized_inventory = characterize(
            dynamic_inventory_df=inventory_in_time_horizon,
            metric=metric,
            characterization_functions=characterization_functions,
            base_lcia_method=self.method,
            time_horizon=time_horizon,
            fixed_time_horizon=fixed_time_horizon,
            time_horizon_start=time_horizon_start,
            characterization_function_co2=characterization_function_co2,
        )

        return self.characterized_inventory

    ###################
    # Core properties #
    ###################

    @property
    def base_score(self) -> float:
        """
        Score of the base LCA, i.e., the "normal" LCA without time-explicit information.
        Same as bw2calc.LCA.score
        """
        return self.base_lca.score

    @property
    def static_score(self) -> float:
        """
        Score resulting from the static LCIA of the time-explicit inventory.
        """
        if not hasattr(self, "lca"):
            raise AttributeError("LCI not yet calculated. Call TimexLCA.lci() first.")
        return self.lca.score

    @property
    def dynamic_score(self) -> float:
        """
        Score resulting from the dynamic LCIA of the time-explicit inventory.
        """
        if not hasattr(self, "characterized_inventory"):
            raise AttributeError(
                "Characterized inventory not yet calculated. Call TimexLCA.dynamic_lcia() first."
            )
        return self.characterized_inventory["amount"].sum()

    ###############################################
    # Other core functions for the inner workings #
    ###############################################

    def build_datapackage(self) -> list:
        """
        Creates the datapackages that contain the modifications to the technosphere and biosphere
        matrix using the `MatrixModifier` class.

        Parameters
        ----------
        None

        Returns
        -------
        list
            List of datapackages that contain the modifications to the technosphere and biosphere
            matrix

        See also
        --------
        bw_timex.matrix_modifier.MatrixModifier: Class that handles the technosphere and biosphere matrix modifications.
        """
        self.matrix_modifier = MatrixModifier(
            self.timeline,
            self.database_dates_static,
            self.demand_timing,
            self.nodes,
            self.interdatabase_activity_mapping,
        )
        self.node_collections["temporal_markets"] = (
            self.matrix_modifier.temporal_market_ids
        )
        self.node_collections["temporalized_processes"] = (
            self.matrix_modifier.temporalized_process_ids
        )
        return self.matrix_modifier.create_datapackage()

    def calculate_dynamic_inventory(
        self,
        expand_technosphere=True,
    ) -> None:
        """
        Calculates the dynamic inventory, by first creating a dynamic biosphere matrix using the
        `DynamicBiosphereBuilder` class and then multiplying it with the dynamic supply array. The
        dynamic inventory matrix is stored in the attribute `dynamic_inventory`. It is also
        converted to a DataFrame and stored in the attribute `dynamic_inventory_df`.

        Parameters
        ----------
        expand_technosphere: bool
           A boolean indicating if the dynamic biosphere matrix is built directly from the expanded
           matrices or from the timeline. Default is True (from expanded matrices).

        Returns
        -------
        None
            calculates the dynamic inventory and stores it in the attribute
            `dynamic_inventory` as a matrix and in `dynamic_inventory_df` as a DataFrame.
            Also calculates and stores the lci of the temporal markets in the attribute
            self.temporal_market_lcis for use in contribution analysis of the background processes.

        See also
        --------
        bw_timex.dynamic_biosphere_builder.DynamicBiosphereBuilder: Class for creating the dynamic biosphere matrix and inventory.
        """

        if not hasattr(self, "lca"):
            raise AttributeError(
                "Time-explicit LCA object does not exist. Call TimexLCA.lci() first."
            )

        self.biosphere_time_mapping = TimeMappingDict(start_id=0)

        self.dynamic_biosphere_builder = DynamicBiosphereBuilder(
            self.lca,
            self.activity_time_mapping,
            self.biosphere_time_mapping,
            self.demand_timing,
            self.node_collections,
            self.temporal_grouping,
            self.database_dates,
            self.database_dates_static,
            self.timeline,
            self.interdatabase_activity_mapping,
            expand_technosphere=expand_technosphere,
        )
        self.dynamic_biosphere_matrix, self.temporal_market_lcis = (
            self.dynamic_biosphere_builder.build_dynamic_biosphere_matrix(
                expand_technosphere=expand_technosphere,
            )
        )

        # Build the dynamic inventory
        count = len(self.dynamic_biosphere_builder.dynamic_supply_array)
        # diagonalization of supply array keeps the dimension of the process, which we want to pass
        # as additional information to the dynamic inventory dict
        diagonal_supply_array = sparse.spdiags(
            [self.dynamic_biosphere_builder.dynamic_supply_array], [0], count, count
        )
        self.dynamic_inventory = self.dynamic_biosphere_matrix @ diagonal_supply_array

        self.dynamic_inventory_df = self.create_dynamic_inventory_dataframe(
            expand_technosphere
        )

    def create_dynamic_inventory_dataframe(
        self,
        expand_technosphere=True,
        use_disaggregated_lci=False,
    ) -> pd.DataFrame:
        """
        Brings the dynamic inventory from its matrix form in `dynamic_inventory` into the
        format of a pandas.DataFrame, with the right structure to later apply dynamic
        characterization functions.

        Format is:

        +------------+--------+------+----------+
        |   date     | amount | flow | activity |
        +============+========+======+==========+
        |  datetime  |   33   |  1   |    2     |
        +------------+--------+------+----------+
        |  datetime  |   32   |  1   |    2     |
        +------------+--------+------+----------+
        |  datetime  |   31   |  1   |    2     |
        +------------+--------+------+----------+

        - date: datetime, e.g. '2024-01-01 00:00:00'
        - flow: flow id
        - activity: activity id

        Parameters
        ----------
        expand_technosphere: bool
            A boolean indicating if the dynamic biosphere matrix is built directly from the
            expanded matrices or from the timeline. Default is True.

        Returns
        -------
        pandas.DataFrame, dynamic inventory in DataFrame format

        """
        if use_disaggregated_lci:
            dynamic_inventory = self.dynamic_inventory_disaggregated
        else:
            dynamic_inventory = self.dynamic_inventory
        dataframe_rows = []
        for i in range(dynamic_inventory.shape[0]):
            row_start = dynamic_inventory.indptr[i]
            row_end = dynamic_inventory.indptr[i + 1]
            for j in range(row_start, row_end):
                row = i
                col = dynamic_inventory.indices[j]
                value = dynamic_inventory.data[j]

                if expand_technosphere:
                    emitting_process_id = self.lca.activity_dict.reversed[col]
                else:
                    emitting_process_id = self.timeline.iloc[col][
                        "time_mapped_producer"
                    ]

                # indices are already the same as in the matrix, as we create an entirely new
                # biosphere instead of adding new entries (like we do with the technosphere matrix)
                bioflow_id, date = self.biosphere_time_mapping.reversed[row]
                dataframe_rows.append(
                    (
                        date,
                        value,
                        bioflow_id,
                        emitting_process_id,
                    )
                )

        df = pd.DataFrame(
            dataframe_rows, columns=["date", "amount", "flow", "activity"]
        )

        df.date = df.date.astype("datetime64[s]")

        return df.sort_values(by=["date", "amount"], ascending=[True, False])

    #############
    # For setup #
    #############

    def prepare_base_lca_inputs(
        self,
        demand=None,
        method=None,
        weighting=None,
        normalization=None,
        demands=None,
        remapping=True,
        demand_database_last=True,
    ) -> tuple:
        """
        Prepare LCA input arguments in Brightway2.5 style.

        Adapted bw2data.compat.py

        The difference to the original method is that we load all available databases into the
        matrices instead of just the ones depending on the demand. We need this for the creation of
        the time mapping dict that creates a mapping between the producer id and the reference
        timing of the databases in the `database_dates`.

        Parameters
        ----------
        demand : dict[object: float]
            The demand for which the LCA will be calculated. The keys can be Brightway `Node`
            instances, `(database, code)` tuples, or integer ids.
        method : tuple
            Tuple defining the LCIA method, such as `('foo', 'bar')`. Only needed if not passing
            `data_objs`.
        weighting : tuple
            Tuple defining the LCIA weighting, such as `('foo', 'bar')`. Only needed if not passing
            `data_objs`.
        normalization: str
        demands: list of dicts of demands
        remapping: bool
            If True, remap dictionaries
        demand_database_last: bool
            If True, add the demand databases last in the list `database_names`.

        Returns
        -------
        tuple
            Indexed demand, data objects, and remapping dictionaries

        See also
        --------
        bw2data.compat.prepare_lca_inputs: Original code this function is adapted from (https://github.com/brightway-lca/brightway2-data/blob/main/bw2data/compat.py).
        """
        if not projects.dataset.data.get("25"):
            raise Brightway2Project(
                "Please use `projects.migrate_project_25` before calculating using Brightway 2.5"
            )

        databases.clean()
        data_objs = []
        remapping_dicts = None

        demand_database_names = list(self.database_dates.keys())

        if demand_database_names:
            database_names = set.union(
                *[
                    Database(db_label).find_graph_dependents()
                    for db_label in demand_database_names
                ]
            )

            if demand_database_last:
                database_names = [
                    x for x in database_names if x not in demand_database_names
                ] + demand_database_names

            data_objs.extend([Database(obj).datapackage() for obj in database_names])

            if remapping:
                # This is technically wrong - we could have more complicated queries
                # to determine what is truly a product, activity, etc.
                # However, for the default database schema, we know that each node
                # has a unique ID, so this won't produce incorrect responses,
                # just too many values. As the dictionary only exists once, this is
                # not really a problem.
                reversed_mapping = {
                    i: (d, c)
                    for d, c, i in AD.select(AD.database, AD.code, AD.id)
                    .where(AD.database << database_names)
                    .tuples()
                }
                remapping_dicts = {
                    "activity": reversed_mapping,
                    "product": reversed_mapping,
                    "biosphere": reversed_mapping,
                }

        if method:
            assert method in methods
            data_objs.append(Method(method).datapackage())
        if weighting:
            assert weighting in weightings
            data_objs.append(Weighting(weighting).datapackage())
        if normalization:
            assert normalization in normalizations
            data_objs.append(Normalization(normalization).datapackage())

        if demands:
            indexed_demand = [{get_id(k): v for k, v in dct.items()} for dct in demands]
        elif demand:
            indexed_demand = {get_id(k): v for k, v in demand.items()}
        else:
            indexed_demand = None

        return indexed_demand, data_objs, remapping_dicts

    def prepare_bw_timex_inputs(
        self,
        demand=None,
        method=None,
        weighting=None,
        normalization=None,
        demands=None,
        remapping=True,
        demand_database_last=True,
    ) -> tuple:
        """
        Prepare LCA input arguments in Brightway 2.5 style.

        ORIGINALLY FROM bw2data.compat.py

        Changes include:
        - always load all databases in demand_database_names
        - indexed_demand has the id of the new consumer_id of the "exploded" demand

        Parameters
        ----------
        demand : dict[object: float]
            The demand for which the LCA will be calculated. The keys can be Brightway `Node`
            instances, `(database, code)` tuples, or integer ids.
        method : tuple
            Tuple defining the LCIA method, such as `('foo', 'bar')`. Only needed if not passing
            `data_objs`.
        demand_timing: dict
            Dictionary mapping demand ids to their timing.
        weighting : tuple
            Tuple defining the LCIA weighting, such as `('foo', 'bar')`. Only needed if not passing
            `data_objs`.
        normalization: str
        demands: list of dicts of demands
        remapping: bool
            If True, remap dictionaries
        demand_database_last: bool
            If True, add the demand databases last in the list `database_names`.

        Returns
        -------
        tuple
            Indexed demand, data objects, and remapping dictionaries

        See also
        --------
        bw2data.compat.prepare_lca_inputs: Original code this function is adapted from (https://github.com/brightway-lca/brightway2-data/blob/main/bw2data/compat.py).
        """

        if not projects.dataset.data.get("25"):
            raise Brightway2Project(
                "Please use `projects.migrate_project_25` before calculating using Brightway 2.5"
            )

        databases.clean()
        data_objs = []
        remapping_dicts = None

        demand_database_names = list(self.database_dates.keys())

        if demand_database_names:
            database_names = set.union(
                *[
                    Database(db_label).find_graph_dependents()
                    for db_label in demand_database_names
                ]
            )

            if demand_database_last:
                database_names = [
                    x for x in database_names if x not in demand_database_names
                ] + demand_database_names

            data_objs.extend([Database(obj).datapackage() for obj in database_names])

            if remapping:
                # This is technically wrong - we could have more complicated queries
                # to determine what is truly a product, activity, etc.
                # However, for the default database schema, we know that each node
                # has a unique ID, so this won't produce incorrect responses,
                # just too many values. As the dictionary only exists once, this is
                # not really a problem.
                reversed_mapping = {
                    i: (d, c)
                    for d, c, i in AD.select(AD.database, AD.code, AD.id)
                    .where(AD.database << database_names)
                    .tuples()
                }
                remapping_dicts = {
                    "activity": reversed_mapping,
                    "product": reversed_mapping,
                    "biosphere": reversed_mapping,
                }

        if method:
            assert method in methods
            data_objs.append(Method(method).datapackage())
        if weighting:
            assert weighting in weightings
            data_objs.append(Weighting(weighting).datapackage())
        if normalization:
            assert normalization in normalizations
            data_objs.append(Normalization(normalization).datapackage())

        if demands:
            indexed_demand = [
                {
                    self.activity_time_mapping[
                        (
                            bd.get_node(id=bd.get_id(k)).key,
                            self.demand_timing[bd.get_id(k)],
                        )
                    ]: v
                    for k, v in dct.items()
                }
                for dct in demands
            ]
        elif demand:
            indexed_demand = {
                self.activity_time_mapping[
                    (
                        ("temporalized", bd.get_node(id=bd.get_id(k))["code"]),
                        self.demand_timing[bd.get_id(k)],
                    )
                ]: v
                for k, v in demand.items()
            }
        else:
            indexed_demand = None

        return indexed_demand, data_objs, remapping_dicts

    def check_format_database_dates(self) -> None:
        """
        Checks that the database_dates is provided by the user in the correct format
        and that the databases are available in the Brightway2 project.

        Parameters
        ----------
        None

        Returns
        -------
        None
            raises an error if the format is not correct or the databases are not available.

        """
        for database, value in self.database_dates.items():
            if not (value == "dynamic" or isinstance(value, datetime)):
                raise ValueError(
                    f"Warning: The timestamp {value} is neither 'dynamic' nor a datetime object. Check the format of the database_dates."
                )

            if database not in bd.databases:
                raise ValueError(
                    f"Database '{database}' not available in the Brightway2 project."
                )

    def create_node_collections(self) -> None:
        """
        Creates a dict of collections of nodes that will be useful down the line, e.g. to determine
        static nodes for the graph traversal or create the dynamic biosphere matrix.
        Available collections are:

        - ``background``: set of node ids of all processes that depend on the demand processes and are in the background databases
        - ``foreground``: set of node ids of all processes that are not in the background databases
        - ``first_level_background_static``: set of node ids of all processes that are in the background databases and are directly linked to the demand processes

        Parameters
        ----------
            None

        Returns
        -------
            None
                adds the `node_collections containing` the above-mentioned collections,
                as well as interdatabase_activity_mapping
        """
        self.node_collections = {}

        # Original variable names preserved, set types for performance and uniqueness
        demand_database_names = {
            db
            for db in self.database_dates.keys()
            if db not in self.database_dates_static.keys()
        }

        demand_dependent_database_names = set()
        demand_dependent_database_names.update(demand_database_names)
        for db in demand_database_names:
            demand_dependent_database_names.update(
                bd.Database(db).find_graph_dependents()
            )

        demand_dependent_background_database_names = (
            demand_dependent_database_names & self.database_dates_static.keys()
        )

        background = {
            node.id
            for db in demand_dependent_background_database_names
            for node in bd.Database(db)
        }
        self.node_collections["background"] = background

        foreground = {
            node.id for db in demand_database_names for node in bd.Database(db)
        }
        self.node_collections["foreground"] = foreground

        first_level_background_static = set()
        foreground_db = bd.Database(list(demand_database_names)[0])
        for node_id in foreground:
            node = foreground_db.get(id=node_id)
            for exc in chain(node.technosphere(), node.substitution()):
                if exc.input["database"] in demand_dependent_background_database_names:
                    first_level_background_static.add(exc.input.id)

        self.node_collections["first_level_background_static"] = (
            first_level_background_static
        )

    def add_interdatabase_activity_mapping_from_timeline(self) -> None:
        """
        Fills the interdatabase_activity_mapping, which is a SetList of the matching processes
        across background databases in the format of {(id, database_name_1), (id, database_name_2)}
        with only those activities and background databases that are actually mapped in the
        timeline.

        Parameters
        ----------
        None


        Returns
        -------
        None
            Adds the ids of producers in other background databases
            (only those interpolated to in the timeline) to the `interdatabase_activity_mapping`.
        """
        if not hasattr(self, "timeline"):
            raise AttributeError(
                "Timeline not yet built. Call TimexLCA.build_timeline() first."
            )

        filtered_timeline = self.timeline.loc[
            self.timeline.temporal_market_shares.notnull()
        ]
        unique_producers = filtered_timeline.producer.unique()

        self.interdatabase_activity_mapping.update(
            {producer: {} for producer in unique_producers}
        )

        producer_tuples_dict = {}
        for producer in unique_producers:
            producer_node = self.nodes[producer]
            producer_tuples_dict[
                (
                    producer_node["name"],
                    producer_node.get("reference product"),
                    producer_node["location"],
                )
            ] = producer

        unique_produces_tuples = producer_tuples_dict.keys()

        for node in self.nodes.values():
            node_tuple = (node["name"], node.get("reference product"), node["location"])
            if node_tuple in unique_produces_tuples:
                producer_id = producer_tuples_dict[node_tuple]
                self.interdatabase_activity_mapping[producer_id][
                    node["database"]
                ] = node.id

        self.interdatabase_activity_mapping.make_reciprocal()

    def collect_temporalized_processes_from_timeline(self) -> None:
        """
        Prepares the input for the LCA from the timeline.

        Parameters
        ----------
        None

        Returns
        -------
        None
            Adds "temporal_markets" and "temporalized_processes" to the
            node_collections based on the timeline.

        """
        unique_producers = (
            self.timeline.groupby(["producer", "time_mapped_producer"])
            .count()
            .index.values
        )

        temporal_market_ids = set()
        temporalized_process_ids = set()

        for producer, time_mapped_producer in unique_producers:
            if self.nodes[producer]["database"] in self.database_dates_static.keys():
                temporal_market_ids.add(time_mapped_producer)
            else:
                temporalized_process_ids.add(time_mapped_producer)

        self.node_collections["temporal_markets"] = temporal_market_ids
        self.node_collections["temporalized_processes"] = temporalized_process_ids

    def add_static_activities_to_activity_time_mapping(self) -> None:
        """
        Adds all activities from the static LCA to `activity_time_mapping`, an instance of
        `TimeMappingDict`. This gives a unique mapping in the form of
        (('database', 'code'), datetime_as_integer): time_mapping_id) that is later used to uniquely
        identify time-resolved processes. Here, the activity_time_mapping is the
        pre-population with the static activities. The time-explicit activities (from other
        temporalized background databases) are added later on by the TimelineBuilder.
        Activities in the foreground database are mapped with
        (('database', 'code'), "dynamic"): time_mapping_id)" as their timing is not yet known.

        Parameters
        ----------
        None

        Returns
        -------
        None
            adds the static activities to the `activity_time_mapping`
        """
        for idx in self.base_lca.dicts.activity.keys():  # activity ids
            key = self.base_lca.remapping_dicts["activity"][idx]  # ('database', 'code')
            time = self.database_dates[
                key[0]
            ]  # datetime (or 'dynamic' for foreground processes)
            if isinstance(time, str):  # if 'dynamic', just add the string
                self.activity_time_mapping.add((key, time), unique_id=idx)
            elif isinstance(time, datetime):
                self.activity_time_mapping.add(
                    (key, extract_date_as_integer(time, self.temporal_grouping)),
                    unique_id=idx,
                )  # if datetime, map to the date as integer
            else:
                raise ValueError(f"Time of activity {key} is neither datetime nor str.")

    def create_demand_timing(self) -> dict:
        """
        Generate a dictionary that maps producer (key) to timing (value) for the demands in the
        product system. It searches the timeline for those rows that contain the functional units
        (demand-processes as producer and -1 as consumer) and returns the time of the demand as an
        integer. Time of demand can have flexible resolution (year=YYYY, month=YYYYMM, day=YYYYMMDD,
        hour=YYYYMMDDHH) defined in `temporal_grouping`.

        Parameters
        ----------
        None

        Returns
        -------
        dict
            Dictionary mapping producer ids to reference timing for the specified demands.
        """
        demand_ids = [bd.get_activity(key).id for key in self.demand.keys()]
        demand_rows = self.timeline[
            self.timeline["producer"].isin(demand_ids)
            & (self.timeline["consumer"] == -1)
        ]
        self.demand_timing = {
            row.producer: row.hash_producer for row in demand_rows.itertuples()
        }
        return self.demand_timing

    ######################################
    # For creating human-friendly output #
    ######################################

    def create_labelled_technosphere_dataframe(self) -> pd.DataFrame:
        """
        Returns the technosphere matrix as a dataframe with comprehensible labels instead of ids.

        Parameters
        ----------
        None

        Returns
        -------
        pd.DataFrame
            technosphere matrix as a pandas.DataFrame with comprehensible labels instead
            of ids.
        """

        df = pd.DataFrame(self.lca.technosphere_matrix.toarray())
        df.rename(  # from matrix id to activity id
            index=self.lca.dicts.activity.reversed,
            columns=self.lca.dicts.activity.reversed,
            inplace=True,
        )
        df.rename(  # from activity id to ((database, code), time)
            index=self.activity_time_mapping.reversed,
            columns=self.activity_time_mapping.reversed,
            inplace=True,
        )
        return df

    def create_labelled_biosphere_dataframe(self) -> pd.DataFrame:
        """
        Returns the biosphere matrix as a pandas.DataFrame with comprehensible labels instead of ids.

        Parameters
        ----------
        None

        Returns
        -------
        pd.DataFrame
            biosphere matrix as a pandas.DataFrame with comprehensible labels instead of
            ids.
        """

        df = pd.DataFrame(self.lca.biosphere_matrix.toarray())
        df.rename(  # from matrix id to activity id
            index=self.lca.dicts.biosphere.reversed,
            columns=self.lca.dicts.activity.reversed,
            inplace=True,
        )
        df.rename(
            index=self.lca.remapping_dicts[
                "biosphere"
            ],  # from activity id to bioflow name
            columns=self.activity_time_mapping.reversed,  # id to ((database, code), time)
            inplace=True,
        )

        return df

    def create_labelled_dynamic_biosphere_dataframe(self) -> pd.DataFrame:
        """
        Returns the dynamic biosphere matrix as a dataframe with comprehensible labels instead of
        ids.

        Parameters
        ----------
        None

        Returns
        -------
        pd.DataFrame
            dynamic biosphere matrix as a pandas.DataFrame with comprehensible labels
            instead of ids.
        """
        df = pd.DataFrame(self.dynamic_biosphere_matrix.toarray())
        df.rename(  # from matrix id to activity id
            index=self.biosphere_time_mapping.reversed,
            columns=self.lca.dicts.activity.reversed,
            inplace=True,
        )
        df.rename(  # from activity id to ((database, code), time)
            columns=self.activity_time_mapping.reversed,
            inplace=True,
        )

        df = df.loc[(df != 0).any(axis=1)]  # For readablity, remove all-zero rows

        return df

    def get_activity_name_from_time_mapped_id(self, time_mapped_id: int) -> str:
        """
        Get the activity name for a time-mapped activity ID.
        Uses the pre-built code-to-name cache for efficient lookups.

        Parameters
        ----------
        time_mapped_id : int
            The time-mapped activity ID from activity_time_mapping

        Returns
        -------
        str
            The name of the activity
        """
        # Extract the code from the activity_time_mapping
        # Structure: time_mapped_id -> (('database', 'code'), time)
        ((_, code), _) = self.activity_time_mapping.reversed[time_mapped_id]

        # Use the pre-built cache for O(1) lookup instead of database query
        return self._activity_code_to_name_cache.get(code, code)

    def create_labelled_dynamic_inventory_dataframe(self) -> pd.DataFrame:
        """
        Returns the dynamic_inventory_df with comprehensible labels for flows and activities instead
        of ids.

        Parameters
        ----------
        None

        Returns
        -------
        pd.DataFrame
            dynamic inventory matrix as a pandas.DataFrame with comprehensible labels
            instead of ids.
        """

        if not hasattr(self, "dynamic_inventory_df"):
            raise AttributeError(
                "Dynamic inventory not yet calculated. Call \
                    TimexLCA.lci(build_dynamic_biosphere=True) first."
            )

        df = self.dynamic_inventory_df.copy()
        df["flow"] = df["flow"].apply(lambda x: bd.get_node(id=x)["name"])

        # Build activity name cache efficiently using pre-built code-to-name cache
        activity_name_cache = {
            activity: self.get_activity_name_from_time_mapped_id(activity)
            for activity in df["activity"].unique()
        }

        df["activity"] = df["activity"].map(activity_name_cache)

        return df

    def plot_dynamic_inventory(self, bio_flows, cumulative=False) -> None:
        """
        Simple plot of dynamic inventory of a biosphere flow over time, with optional cumulative
        plotting.

        Parameters
        ----------
        bio_flows : list of int
            database ids of the biosphere flows to plot.
        cumulative : bool
            if True, plot cumulative amounts over time

        Returns
        -------
        None
            shows a plot
        """
        plt.figure(figsize=(14, 6))

        filtered_df = self.dynamic_inventory_df[
            self.dynamic_inventory_df["flow"].isin(bio_flows)
        ]
        aggregated_df = filtered_df.groupby("date").sum()["amount"].reset_index()

        if cumulative:
            aggregated_df["amount"] = np.cumsum(aggregated_df["amount"])

        plt.plot(
            aggregated_df["date"], aggregated_df["amount"], marker="o", linestyle="none"
        )

        plt.ylim(bottom=0)
        plt.xlabel("time")
        plt.ylabel("amount [kg]")
        plt.grid(True)
        plt.tight_layout()  # Adjust layout to make room for the rotated date labels
        plt.show()

    def plot_dynamic_characterized_inventory(
        self,
        cumsum: bool = False,
        sum_emissions_within_activity: bool = False,
        sum_activities: bool = False,
    ) -> None:
        """
        Plot the characterized inventory of the dynamic LCI in a very simple plot.
        Legend and title are selected automatically based on the chosen metric.

        Parameters
        ----------
        cumsum : bool
            if True, plot cumulative amounts over time
        sum_emissions_within_activity : bool
            if True, sum emissions within each activity over time
        sum_activities : bool
            if True, sum emissions over all activities over time

        Returns
        -------
        None
            shows a plot
        """

        if not hasattr(self, "characterized_inventory"):
            raise AttributeError(
                "Characterized inventory not yet calculated. Call TimexLCA.dynamic_lcia() first."
            )

        metric_ylabels = {
            "radiative_forcing": "radiative forcing [W/m²]",
            "GWP": f"GWP{self.current_time_horizon} [kg CO₂-eq]",
        }

        # Fetch the inventory to use in plotting, modify based on flags
        plot_data = self.characterized_inventory.copy()

        if cumsum:
            plot_data["amount_sum"] = plot_data["amount"].cumsum()
            amount = "amount_sum"
        else:
            amount = "amount"

        if sum_emissions_within_activity:
            plot_data = plot_data.groupby(["date", "activity"]).sum().reset_index()
            plot_data["amount_sum"] = plot_data["amount"].cumsum()

        if sum_activities:
            plot_data = plot_data.groupby("date").sum().reset_index()
            plot_data["amount_sum"] = plot_data["amount"].cumsum()
            plot_data["activity_label"] = "All activities"

        else:  # plotting activities separate
            # Build activity name cache efficiently using pre-built code-to-name cache
            activity_name_cache = {
                activity: self.get_activity_name_from_time_mapped_id(activity)
                for activity in plot_data["activity"].unique()
            }

            plot_data["activity_label"] = plot_data["activity"].map(activity_name_cache)

        # Plotting
        plt.figure(figsize=(14, 6))
        axes = sb.scatterplot(x="date", y=amount, hue="activity_label", data=plot_data)

        # Determine y-axis limit flexibly
        if plot_data[amount].min() < 0:
            ymin = plot_data[amount].min() * 1.1
        else:
            ymin = 0

        axes.set_axisbelow(True)
        axes.set_ylim(bottom=ymin)
        axes.set_ylabel(metric_ylabels[self.current_metric])
        axes.set_xlabel("time")

        handles, labels = axes.get_legend_handles_labels()
        axes.legend(handles[::-1], labels[::-1])
        plt.grid(True)
        plt.show()



from datetime import datetime

import bw2data as bd
import pytest

from bw_timex import TimexLCA

# from .fixtures.substitution_db_fixture import substitution_db_level1, substitution_db_level2


# make sure the test db is loaded
def test_db_fixture(substitution_db):
    assert len(bd.databases) == 4


# for now, one test class, but could be set up more modularly
@pytest.mark.usefixtures("substitution_db")
class TestClass_substitution:

    @pytest.fixture(autouse=True)
    def setup_method(self):

        self.node_a = bd.get_node(database="foreground", code="A")

        database_dates = {
            "db_2020": datetime.strptime("2020", "%Y"),
            "db_2030": datetime.strptime("2030", "%Y"),
            "foreground": "dynamic",
        }

        self.tlca = TimexLCA(
            demand={self.node_a.key: 1},
            method=("GWP", "example"),
            database_dates=database_dates,
        )

        self.tlca.build_timeline(starting_datetime="2024-01-01")
        self.tlca.lci()
        self.tlca.static_lcia()

    def test_substitution(self):
        expected_substitution_score = (
            1
            + 1 * -0.75 * 0.5  # direct emissions at A
            + 5  # substituted emissions from Sub via A in 2020 from db_2020
            * -1
            * 0.7
            * 0.8
            + 5  # substituted emissions from Sub  via B in 2028 from db_2030
            * -1
            * 0.5
            * 0.2
        )  # substituted emissions from Sub bia B in 2028 from db_2020

        assert self.tlca.static_score == pytest.approx(
            expected_substitution_score, rel=0.0001
        )



"""
Testing if the temporal grouping options (year, month, day and hour) work correctly.
"""

from datetime import datetime
from pathlib import Path

import bw2calc as bc
import bw2data as bd
import numpy as np
import pytest
from bw_temporalis import TemporalDistribution

from bw_timex import TimexLCA


@pytest.mark.usefixtures("temporal_grouping_db_monthly")
def test_monthly_resolution_score():

    fu = bd.get_node(database="foreground", code="A")

    database_dates = {
        "db_2022": datetime.strptime("2022", "%Y"),
        "db_2024": datetime.strptime("2024", "%Y"),
        "foreground": "dynamic",
    }

    tlca = TimexLCA(
        demand={fu.key: 1},
        method=("GWP", "example"),
        database_dates=database_dates,
    )
    tlca.build_timeline(
        temporal_grouping="month",
        starting_datetime=datetime.strptime("2024-01-02", "%Y-%m-%d"),
    )
    tlca.lci()
    tlca.static_lcia()
    print(tlca.timeline)

    expected_score = 1 / 3 * 15 + 1 / 6 * 15 + 1 / 6 * 10 + 1 / 3 * 10

    assert tlca.static_score == expected_score


@pytest.mark.usefixtures("temporal_grouping_db_daily")
def test_daily_resolution_score():

    # first change TD in database to daily (although not strictly necessary)
    fu = bd.get_node(database="foreground", code="A")

    database_dates = {
        "db_2022": datetime.strptime(
            "2024-01-10", "%Y-%m-%d"
        ),  # give dbs also a daily resolution (makes calculating the expected score easier)
        "db_2024": datetime.strptime("2024-01-25", "%Y-%m-%d"),
        "foreground": "dynamic",
    }

    tlca = TimexLCA(
        demand={fu.key: 1},
        method=("GWP", "example"),
        database_dates=database_dates,
    )
    tlca.build_timeline(
        temporal_grouping="day",
        starting_datetime=datetime.strptime("2024-01-25", "%Y-%m-%d"),
    )
    tlca.lci()
    tlca.static_lcia()
    print(tlca.timeline)

    expected_score = 1 / 3 * 15 + 2 / 9 * 15 + 1 / 9 * 10 + 1 / 3 * 10

    assert tlca.static_score == pytest.approx(expected_score, rel=0.00005)


@pytest.mark.usefixtures("temporal_grouping_db_hourly")
def test_hourly_resolution_score():

    # first change TD in database to hourly (although not strictly necessary)
    fu = bd.get_node(database="foreground", code="A")

    database_dates = {
        "db_2022": datetime.strptime(
            "2024-01-10 06:00", "%Y-%m-%d %H:%M"
        ),  # give dbs also a hourly resolution (makes calculating the expected score easier)
        "db_2024": datetime.strptime("2024-01-10 21:00", "%Y-%m-%d %H:%M"),
        "foreground": "dynamic",
    }

    tlca = TimexLCA(
        demand={fu.key: 1},
        method=("GWP", "example"),
        database_dates=database_dates,
    )
    tlca.build_timeline(
        temporal_grouping="hour",
        starting_datetime=datetime.strptime("2024-01-10 21:00", "%Y-%m-%d %H:%M"),
    )
    tlca.lci()
    tlca.static_lcia()
    print(tlca.timeline)

    expected_score = 1 / 3 * 15 + 2 / 9 * 15 + 1 / 9 * 10 + 1 / 3 * 10

    assert tlca.static_score == pytest.approx(expected_score, rel=0.00005)



from datetime import datetime, timedelta
from typing import Callable, KeysView, Union

import bw2data as bd
import numpy as np
import pandas as pd
from bw2calc import LCA
from bw2data.configuration import labels
from loguru import logger

from .edge_extractor import Edge, EdgeExtractor
from .utils import (
    convert_date_string_to_datetime,
    extract_date_as_integer,
    extract_date_as_string,
    round_datetime,
)


class TimelineBuilder:
    """
    Class for building a process timeline based on the temporal distributions of their exchanges.

    First, the `EdgeExtractor` does a priority-first graph traversal and extracts a timeline of
    exchanges (edge_timeline) with temporal information. Identical edges within temporal grouping
    (e.g. year, month, day, hour) are then grouped and the amount of exchanges is summed up.
    """

    def __init__(
        self,
        base_lca: LCA,
        starting_datetime: datetime,
        edge_filter_function: Callable,
        database_dates: dict,
        database_dates_static: dict,
        activity_time_mapping: dict,
        node_collections: dict,
        nodes: dict,
        temporal_grouping: str = "year",
        interpolation_type: str = "linear",
        cutoff: float = 1e-9,
        max_calc: int = 2000,
        *args,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        base_lca: LCA
            A static LCA object.
        starting_datetime: datetime | str, optional
            Point in time when the demand occurs.
        edge_filter_function: Callable
            A callable that filters edges. If not provided, a function that always returns False is used.
        database_dates: dict
            A dictionary mapping databases to dates.
        database_dates_static: dict
            same as database_dates, but excluding the "dynamic" foreground databases.
        activity_time_mapping: dict
          A dictionary to map processes to specific times.
        node_collections: dict
            A dictionary collecting useful subsets of node ids.
        nodes: dict
            A dictionary {node_id: 'bw2data.backends.proxies.Activity'} for all nodes.
        temporal_grouping: str, optional
            The temporal grouping to be used. Default is "year".
        interpolation_type: str, optional
            The type of interpolation to be used to select the background databases. Default is "linear".
        cutoff:
            The cutoff value for the graph traversal. Default is 1e-9.
        max_calc:
            The maximum number of calculations to be performed by the graph traversal. Default is 2000.
        args:   Variable length argument list
            Keyword arguments passed to the EdgeExtractor which inherits from TemporalisLCA. Here, things like the further settings for graph traversal can be set. For details, see bw_temporalis.TemporalisLCA.
        kwargs: Arbitrary keyword arguments
            Keyword arguments passed to the EdgeExtractor which inherits from TemporalisLCA.
        """
        self.base_lca = base_lca
        self.starting_datetime = starting_datetime
        self.edge_filter_function = edge_filter_function
        self.database_dates = database_dates
        self.database_dates_static = database_dates_static
        self.activity_time_mapping = activity_time_mapping
        self.node_collections = node_collections
        self.nodes = nodes
        self.temporal_grouping = temporal_grouping
        self.interpolation_type = interpolation_type
        self.cutoff = cutoff
        self.max_calc = max_calc

        # Finding indices of activities from the connected background databases that are known to be static, i.e. have no temporal distributions connecting to them.
        # These will be be skipped in the graph traversal.
        static_background_activity_ids = {
            node_id
            for node_id in self.node_collections["background"]
            if node_id not in self.node_collections["first_level_background_static"]
        }

        logger.info("Traversing supply chain graph...")
        self.edge_extractor = EdgeExtractor(
            base_lca,
            starting_datetime=self.starting_datetime,
            *args,
            edge_filter_function=edge_filter_function,
            cutoff=self.cutoff,
            max_calc=self.max_calc,
            static_activity_indices=set(static_background_activity_ids),
            **kwargs,
        )
        self.edge_timeline = self.edge_extractor.build_edge_timeline()

    def build_timeline(self) -> pd.DataFrame:
        """
        Create a DataFrame with grouped, time-explicit edges and, for each grouped edge,
        interpolate to the database with the closest time of representativeness.

        It uses the edge_timeline, an output from the graph traversal in EdgeExtractor.
        Edges from same producer to same consumer that occur at different times within
        the same time window (temporal_grouping) are grouped together.
        Possible temporal groupings are "year", "month", "day" and "hour".

        For edges between foreground and background system, the column "temporal_market_shares"
        assigns the ratio [0-1] of the edge's amount to be taken from the database with the closest
        time of representativeness. If a process is in the foreground system only, the interpolation weight is set to None.

        Available interpolation types are:

        - "linear": linear interpolation between the two closest databases, based on temporal distance.

        - "closest": closest database is assigned 1

        Parameters
        ----------
        None
            (all are already passed during instantiation)

        Returns
        -------
        pd.DataFrame
            A timeline with grouped, time-explicit edges and temporal_market_shares to background databases.
        """
        logger.info("Building timeline...")

        # check if database names match with databases in BW project
        self.check_database_names()

        # Check if temporal_grouping is a valid value
        valid_temporal_groupings = ["year", "month", "day", "hour"]
        if self.temporal_grouping not in valid_temporal_groupings:
            raise ValueError(
                f"Invalid value for 'temporal_grouping'. Allowed values are {valid_temporal_groupings}."
            )

        # Extract edge data into a list of dictionaries
        edges_data = [self.extract_edge_data(edge) for edge in self.edge_timeline]

        # Convert list of dictionaries to dataframe
        edges_df = pd.DataFrame(edges_data)

        # adjust the sign for substitution exchanges:
        edges_df["amount"] = edges_df["amount"] * edges_df["edge_type"].apply(
            lambda x: self.adjust_sign_of_amount_based_on_edge_type(x)
        )

        # Explode datetime and amount columns: each row with multiple dates and amounts is exploded into multiple rows with one date and one amount
        edges_df = edges_df.explode(["consumer_date", "producer_date", "amount"])
        edges_df.drop_duplicates(inplace=True)
        edges_df = edges_df[edges_df["amount"] != 0]

        # For the Functional Unit: set consumer date = producer date as it occurs at the same time
        edges_df.loc[edges_df["consumer"] == -1, "consumer_date"] = edges_df.loc[
            edges_df["consumer"] == -1, "producer_date"
        ]

        edges_df["rounded_consumer_date"] = edges_df["consumer_date"].apply(
            lambda x: round_datetime(x, self.temporal_grouping)
        )
        edges_df["rounded_producer_date"] = edges_df["producer_date"].apply(
            lambda x: round_datetime(x, self.temporal_grouping)
        )

        # extract grouping time of consumer and producer: processes occuring at different times within in the same time window of grouping get the same grouping time
        edges_df["consumer_grouping_time"] = edges_df["rounded_consumer_date"].apply(
            lambda x: extract_date_as_string(x, self.temporal_grouping)
        )
        edges_df["producer_grouping_time"] = edges_df["rounded_producer_date"].apply(
            lambda x: extract_date_as_string(x, self.temporal_grouping)
        )

        # group unique pair of consumer and producer with the same grouping times
        grouped_edges = (
            edges_df.groupby(
                [
                    "producer_grouping_time",
                    "consumer_grouping_time",
                    "producer",
                    "consumer",
                ]
            )
            .agg({"amount": "sum"})
            .reset_index()
        )

        # convert grouping times, which was only used as intermediate variable, back to datetime
        grouped_edges["date_producer"] = grouped_edges["producer_grouping_time"].apply(
            lambda x: convert_date_string_to_datetime(self.temporal_grouping, x)
        )
        grouped_edges["date_consumer"] = grouped_edges["consumer_grouping_time"].apply(
            lambda x: convert_date_string_to_datetime(self.temporal_grouping, x)
        )

        # add dates as integers as hashes to the DataFrame
        grouped_edges["hash_producer"] = grouped_edges["date_producer"].apply(
            lambda x: extract_date_as_integer(x, time_res=self.temporal_grouping)
        )

        grouped_edges["hash_consumer"] = grouped_edges["date_consumer"].apply(
            lambda x: extract_date_as_integer(x, time_res=self.temporal_grouping)
        )

        # add new processes to activity_time_mapping
        for row in grouped_edges.itertuples():
            self.activity_time_mapping.add(
                (
                    ("temporalized", self.nodes[row.producer]["code"]),
                    row.hash_producer,
                )
            )

        # store the ids from the time_mapping in DataFrame
        grouped_edges["time_mapped_producer"] = grouped_edges.apply(
            lambda row: self.get_time_mapping_key(row.producer, row.hash_producer),
            axis=1,
        )

        grouped_edges["time_mapped_consumer"] = grouped_edges.apply(
            lambda row: (
                self.get_time_mapping_key(row.consumer, row.hash_consumer)
                if row.consumer != -1
                else -1
            ),
            axis=1,
        )

        # Add temporal_market_shares to background databases to the DataFrame
        grouped_edges = self.add_column_temporal_market_shares_to_timeline(
            grouped_edges,
            interpolation_type=self.interpolation_type,
        )

        # Retrieve producer and consumer names
        grouped_edges["producer_name"] = grouped_edges.producer.apply(
            lambda x: self.nodes[x]["name"]
        )
        grouped_edges["consumer_name"] = grouped_edges.consumer.apply(
            self.get_consumer_name
        )

        # Reorder columns
        grouped_edges = grouped_edges[
            [
                "hash_producer",
                "time_mapped_producer",
                "date_producer",
                "producer",
                "producer_name",
                "hash_consumer",
                "time_mapped_consumer",
                "date_consumer",
                "consumer",
                "consumer_name",
                "amount",
                "temporal_market_shares",
            ]
        ]

        return grouped_edges

    ###################################################
    # underlying functions called by build_timeline() #
    ###################################################

    def check_database_names(self) -> None:
        """
        Check that the strings of the databases exist in the databases of the Brightway project.

        """
        for db in self.database_dates_static.keys():
            assert (
                db in bd.databases
            ), f"{db} is not in your Brightway project databases."

    def extract_edge_data(self, edge: Edge) -> dict:
        """
        Stores the attributes of an Edge instance in a dictionary.

        Parameters
        ----------
        edge: Edge
            Edge instance

        Returns
        -------
        dict
            Dictionary with the attributes of the edge instance.
        """
        try:
            consumer_date = edge.abs_td_consumer.date
            consumer_date = np.array(
                [consumer_date for i in range(len(edge.td_producer))]
            ).T.flatten()
        except AttributeError:
            consumer_date = None

        return {
            "producer": edge.producer,
            "consumer": edge.consumer,
            "consumer_date": consumer_date,
            "producer_date": edge.abs_td_producer.date,
            "amount": edge.abs_td_producer.amount,
            "edge_type": edge.edge_type,
        }

    def adjust_sign_of_amount_based_on_edge_type(self, edge_type):
        """
        It checks if the an exchange is of type substitution or a technosphere exchange,
        based on bw2data labelling convention, and adjusts the amount accordingly.
        Flips the sign of the amount value in the timeline for substitution (positive technosphere) exchanges.

        Parameters
        ----------
        edge_type: str
            Type of the edge, as defined in the exchange data.

        Returns
        -------
        int
            Multiplier for the amount value, 1 for technosphere exchanges, -1 for substitution exchanges.

        """

        if edge_type in labels.technosphere_negative_edge_types:
            return 1  # Variants of technosphere labels

        if edge_type in labels.technosphere_positive_edge_types:
            return -1 if edge_type in labels.substitution_edge_types else 1

        raise TypeError(f"Unrecognized type in this edge: {edge_type}")

    def get_time_mapping_key(self, node_id: int, node_hash: int) -> int:
        """
        Returns the time_mapping_id (key) from the activity_time_mapping for a given node.

        Parameters
        ----------
        node_id: int
            database id of the node.
        node_hash: int
            datetime_as_integer of the node.

        Returns
        -------
        int
            time_mapping_id (key) of the corresponding time-mapped activity.

        """
        try:
            return self.activity_time_mapping[
                (("temporalized", self.nodes[node_id]["code"]), node_hash)
            ]
        except KeyError:
            return self.activity_time_mapping[((self.nodes[node_id].key), node_hash)]

    def add_column_temporal_market_shares_to_timeline(
        self,
        tl_df: pd.DataFrame,
        interpolation_type: str = "linear",
    ) -> pd.DataFrame:
        """
        Add a column to a timeline with the weights for an interpolation between
        the two nearest dates, from the list of dates of the available databases.

        Parameters
        ----------
        tl_df: pd.DataFrame
            Timeline as a DataFrame.
        interpolation_type: str, optional
            Type of interpolation between the nearest lower and higher dates.
            Available options: "linear" and "nearest", defaulting to "linear".

        Returns
        -------
        pd.DataFrame
            Timeline as a DataFrame with a column 'temporal_market_shares' added,
            this column looks like {database_name: weight, database_name: weight}.
        """
        if not self.database_dates_static:
            tl_df["temporal_market_shares"] = None
            logger.info(
                "No time-explicit databases are provided. Mapping to time-explicit databases is not possible.",
            )
            return tl_df

        dates_list = [
            date
            for date in self.database_dates_static.values()
            if isinstance(date, datetime)
        ]
        if "date_producer" not in list(tl_df.columns):
            raise ValueError("The timeline does not contain dates.")

        # create reversed dict {date: database} with only static "background" db's
        self.reversed_database_dates = {
            v: k
            for k, v in self.database_dates_static.items()
            if isinstance(v, datetime)
        }

        if self.interpolation_type == "nearest":
            tl_df["temporal_market_shares"] = tl_df["date_producer"].apply(
                lambda x: self.find_closest_date(x, dates_list)
            )

        if self.interpolation_type == "linear":
            tl_df["temporal_market_shares"] = tl_df["date_producer"].apply(
                lambda x: self.get_weights_for_interpolation_between_nearest_years(
                    x, dates_list, interpolation_type
                )
            )

        else:
            raise ValueError(
                f"Sorry, but {interpolation_type} interpolation is not available yet."
            )

        tl_df["temporal_market_shares"] = tl_df.apply(
            self.add_interpolation_weights_at_intersection_to_background, axis=1
        )  # add the weights to the timeline for processes at intersection

        return tl_df

    def find_closest_date(self, target: datetime, dates: KeysView[datetime]) -> dict:
        """
        Find the closest date to the target in the dates list.

        Parameters
        ----------
        target : datetime.datetime
            Target datetime object.
        dates : KeysView[datetime]
            List of datetime.datetime objects.

        Returns
        -------
        dict
            Dictionary with the key as the closest datetime.datetime object from the list and a value of 1.
        """

        # If the list is empty, return None
        if not dates:
            return None

        # Sort the dates
        dates = sorted(dates)

        # Use min function with a key based on the absolute difference between the target and each date
        closest = min(dates, key=lambda date: abs(target - date))

        return {closest: 1}

    def get_weights_for_interpolation_between_nearest_years(
        self,
        reference_date: datetime,
        dates_list: KeysView[datetime],
        interpolation_type: str = "linear",
    ) -> dict:
        """
        Find the nearest dates (lower and higher) for a given date from a list of dates
        and calculate the interpolation weights based on temporal proximity.

        Parameters
        ----------
        reference_date : datetime
            Target date.
        dates_list : KeysView[datetime]
            List of datetime objects representing the temporal representativeness of the available databases.
        interpolation_type : str, optional
            Type of interpolation between the nearest lower and higher dates. For now, only "linear" is available.

        Returns
        -------
        dict
            Dictionary with datetimes of the available closest databases as keys and the weights for interpolation as values.
        """
        dates_list = sorted(dates_list)

        diff_dates_list = [reference_date - x for x in dates_list]

        if timedelta(0) in diff_dates_list:  # date of process == date of database
            exact_match = dates_list[diff_dates_list.index(timedelta(0))]
            return {exact_match: 1}

        closest_lower = None
        closest_higher = None

        # select the closest lower and higher dates of the database in regards to the date of process
        for date in dates_list:
            if date < reference_date:
                if (
                    closest_lower is None
                    or reference_date - date < reference_date - closest_lower
                ):
                    closest_lower = date
            elif date > reference_date:
                if (
                    closest_higher is None
                    or date - reference_date < closest_higher - reference_date
                ):
                    closest_higher = date

        if closest_lower is None:
            logger.info(
                f"Reference date {reference_date} is lower than all provided dates. Data will be taken from the closest higher year.",
            )
            return {closest_higher: 1}

        if closest_higher is None:
            logger.info(
                f"Reference date {reference_date} is higher than all provided dates. Data will be taken from the closest lower year.",
            )
            return {closest_lower: 1}

        if self.interpolation_type == "linear":
            weight = int((reference_date - closest_lower).total_seconds()) / int(
                (closest_higher - closest_lower).total_seconds()
            )
        else:
            raise ValueError(
                f"Sorry, but {interpolation_type} interpolation is not available yet."
            )
        return {closest_lower: round(1 - weight, 3), closest_higher: round(weight, 3)}

    def add_interpolation_weights_at_intersection_to_background(
        self, row
    ) -> Union[dict, None]:
        """
        returns the interpolation weights to background databases only for those exchanges,
        where the producing process actually comes from a background database (temporal markets).

        Only these processes are receiving inputs from the background databases.
        All other process in the timeline are not directly linked to the background,
        so the interpolation weight info is not needed and set to None

        Parameters
        ----------
        row : pd.Series
            Row of the timeline DataFrame
        Returns
        -------
        dict
            Dictionary with the name of databases and interpolation weights.
        """

        if row["producer"] in self.node_collections["first_level_background_static"]:
            return {
                self.reversed_database_dates[x]: v
                for x, v in row["temporal_market_shares"].items()
            }
        return None

    def get_consumer_name(self, idx: int) -> str:
        """
        Returns the name of consumer node.
        If consuming node is the functional unit, returns -1.

        Parameters
        ----------
        id : int
            Id of node.

        Returns
        -------
        str
            Name of the node or -1
        """
        try:
            return self.nodes[idx]["name"]
        except KeyError:
            return "-1"  # functional unit



from datetime import datetime

import bw2data as bd
import numpy as np
import pandas as pd
import pytest

from bw_timex import TimexLCA


# make sure the test db is loaded
def test_dynamic_biosphere_matrix_fixture(dynamic_biosphere_matrix_db):
    assert len(bd.databases) == 3


@pytest.mark.usefixtures("dynamic_biosphere_matrix_db")
class TestClass_dynamic_biosphere_matrix:

    @pytest.fixture(autouse=True)
    def setup_method(self):
        self.node_a = bd.get_node(database="foreground", code="A")

        database_dates = {
            "db_2020": datetime.strptime("2020", "%Y"),
            "foreground": "dynamic",
        }

        self.tlca = TimexLCA(
            demand={self.node_a: 1},
            method=("GWP", "example"),
            database_dates=database_dates,
        )

        self.tlca.build_timeline(
            starting_datetime=datetime.strptime("2024-01-02", "%Y-%m-%d")
        )
        self.tlca.lci(expand_technosphere=True, build_dynamic_biosphere=True)
        self.tlca.static_lcia()

    def test_identical_time_mapped_producers_exist_in_timeline(self):

        duplicates = self.tlca.timeline["time_mapped_producer"].duplicated().any()
        assert duplicates, "No duplicates found in column time_mapped_producer"

    def test_dynamic_biosphere_matrix_for_multiple_identical_time_mapped_producers_in_timeline(
        self,
    ):

        expected_dynamic_biosphere_matrix_as_array = np.array(
            [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.5]]
        )  # 1.5 CO2 is only added once for the time_mapped producer, depsite it occuring twice in the timeline! Before #78, it was added twice , resulting in 3.0

        assert np.array_equal(
            self.tlca.dynamic_biosphere_matrix.toarray(),
            expected_dynamic_biosphere_matrix_as_array,
        ), "Arrays are not equal"



import warnings

from loguru import logger


class SetList:
    """A helper class for the mapping of the same/mapped activity in different (temporal) databases, composed of a list of sets, that hold can hold the set of tuples of (act_id, database).
    It is built by adding sets to the list, and returns the matching sets if one
    calls the an item from a set.

    Example: If the class instance is called my_setlist, my_setlist.add(set).
    """

    def __init__(
        self,
    ) -> None:
        self.list = []

    def add(
        self,
        new_set: set,
    ) -> None:
        """
        This method adds a set to the SetList instance.

        Parameters
        -----
        new_set: a set to add to the SetList instance

        Returns
        -------
        None

        """

        if new_set not in self.list:
            checklist_items = [
                item for itemset in self.list for item in new_set if item in itemset
            ]
            checklist_sets = [
                itemset for itemset in self.list for item in new_set if item in itemset
            ]
            if len(checklist_items) != 0:
                logger.info(
                    f"tried to add {new_set} to the SetList\n, but {checklist_items} already exist in the SetList in:\n {checklist_sets}. \n Skipping {new_set}"
                )
            else:
                self.list.append(new_set)
        else:
            pass

    def __getitem__(self, key: any) -> set:
        """
        Returns all sets in the SetList instance containing the key
        Inputs
        ------
        key: the key to look for in the sets of the SetList

        Returns
        -------
        A list containing the set or all sets
        """
        sets = [matching_set for matching_set in self.list if key in matching_set]
        if len(sets) > 1:
            logger.warning(
                f"Key found in multiple sets! Please check {sets} ! Returning only the first set"
            )
        if len(sets) > 0:
            return sets[0]

        warnings.warn(f"Key {key} not found in SetList")
        return None

    def __len__(
        self,
    ) -> int:
        return len(self.list)

    def __repr__(self):
        return f"SetList({self.list})"


class TimeMappingDict(dict):
    """
    A dictionary mapping (flow, timestamp) tuples to unique integer IDs.
    """

    def __init__(self, start_id=2, *args, **kwargs) -> None:
        """Initializes the dictionary with a starting ID."""
        super().__init__(*args, **kwargs)
        self._current_id = start_id
        self._used_ids = set(self.values())  # Track assigned IDs efficiently
        self._modified = False  # Track changes
        self._reversed_dict = None  # Store reversed dict when needed

    def add(self, process_time_tuple, unique_id=None):
        """
        Adds a new process_time_tuple to the dictionary.

        Parameters
        ----------
        process_time_tuple : tuple
            A tuple of (flow and timestamp)
        unique_id : int, optional
            A unique ID for the tuple (default: None).

        Returns
        -------
        int
            The assigned unique ID.
        """
        # If already exists, return immediately
        if process_time_tuple in self:
            return self[process_time_tuple]

        if unique_id is not None:
            if unique_id in self._used_ids:
                raise ValueError(f"Unique ID {unique_id} is already assigned.")
            self._used_ids.add(unique_id)
            self[process_time_tuple] = unique_id
        else:
            new_id = self._current_id
            self._current_id += 1
            self[process_time_tuple] = new_id
            self._used_ids.add(new_id)

        self._modified = True  # Mark as modified
        return self[process_time_tuple]

    @property
    def reversed(self):
        """
        Returns a reversed version of the dictionary, updating it only if necessary.

        Returns
        -------
        dict
            A reversed dictionary mapping unique IDs to (flow, timestamp) tuples.
        """
        if self._modified or self._reversed_dict is None:
            self._reversed_dict = {v: k for k, v in self.items()}
            self._modified = False  # Reset modification flag
        return self._reversed_dict


class InterDatabaseMapping(dict):
    """
    A dictionary of the form {id1:{database1: id1, database2: id2, ...}, id2: ...} that maps the
    same activity in different databases.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._reciprocal = False

    def find_match(self, id_, db_name) -> any:
        """
        Returns the matching ID in the given database for the given ID.
        """
        return self[id_][db_name]

    def make_reciprocal(self):
        """
        Internal method to make the mapping reciprocal.
        """
        if not self._reciprocal:
            for mapping in list(self.values()):
                for id_ in list(mapping.values()):
                    if id_ not in self.keys():
                        self[id_] = mapping
            self._reciprocal = True

    def __getitem__(self, key):
        if not self._reciprocal:
            self.make_reciprocal()
        return super().__getitem__(key)



import uuid
from typing import Optional

import bw2data as bd
import bw_processing as bwp
import numpy as np
import pandas as pd

from .helper_classes import InterDatabaseMapping


class MatrixModifier:
    """
    Class for adding and re-linking time-explicit processes in the LCA matrices.

    This is done by creating datapackages that add or change matrix entries in the technosphere and
    biosphere matrices, based on a process timeline (from TimelineBuilder.build_timeline()).
    "Temporal markets" are created for processes that are linked to background databases, temporally
    distributing the amounts to time-explicit background databases.
    """

    def __init__(
        self,
        timeline: pd.DataFrame,
        database_dates_static: dict,
        demand_timing: dict,
        nodes: dict,
        interdatabase_activity_mapping: InterDatabaseMapping,
        name: Optional[str] = None,
    ) -> None:
        """
        Initializes the MatrixModifier object and creates empty sets to collect
        the ids of temporalized processes and temporal markets.

        Parameters
        ----------
        timeline : pd.DataFrame
            A DataFrame of the timeline of exchanges
        database_dates_static : dict
            A dictionary mapping the static background databases to dates.
        demand_timing : dict
            A dictionary mapping the demand to its timing.
        name : str, optional
            An optional name for the MatrixModifier instance. Default is None.

        """

        self.timeline = timeline
        self.database_dates_static = database_dates_static
        self.demand_timing = demand_timing
        self.nodes = nodes
        self.interdatabase_activity_mapping = interdatabase_activity_mapping
        self.name = name
        self.temporalized_process_ids = set()
        self.temporal_market_ids = set()

    def create_datapackage(self) -> None:
        """
        Creates a list of datapackages for the technosphere and biosphere matrices,
        by calling the respective functions.

        Parameters
        ----------
        None

        Returns
        -------
        list
            A list of the technosphere and biosphere datapackages.
        """
        technosphere_datapackage = self.create_technosphere_datapackage()
        biosphere_datapackage = self.create_biosphere_datapackage()

        return [technosphere_datapackage, biosphere_datapackage]

    def create_technosphere_datapackage(self) -> bwp.Datapackage:
        """
        Creates the modifications to the technosphere matrix in form of a datapackage.
        Datapackages add or overwrite data points in the LCA matrices before LCA calculations.
        The technosphere datapackage adds the temporalized processes from the timeline
        to the technosphere matrix.

        The heavy lifting of this method happens in the method `add_row_to_technosphere_datapackage()`.
        Here, each node with a temporal distribution is "exploded", which means each occurrence
        of this node (e.g. steel production on 2020-01-01 and steel production on 2015-01-01)
        becomes a separate, time-explicit new node, by adding the new elements to the technosphere matrix.
        For processes at the intersection with background databases, the timing of the exchanges
        determines which background database to link to in so called "Temporal Markets".

        Parameters
        ----------
        None

        Returns
        -------
        bwp.Datapackage
            A datapackage containing the modifications for the technosphere matrix.
        """
        datapackage_technosphere = bwp.create_datapackage(
            sum_inter_duplicates=False
        )  # 'sum_inter_duplicates=False': If the same market is used by multiple foreground processes, the market gets created again, inputs should not be summed.

        new_nodes = set()

        for row in self.timeline.iloc[::-1].itertuples():
            self.add_row_to_technosphere_datapackage(
                row,
                datapackage_technosphere,
                new_nodes,
            )

        # Adding the production exchanges for new nodes
        for node_id, production_amount in new_nodes:
            datapackage_technosphere.add_persistent_vector(
                matrix="technosphere_matrix",
                name=uuid.uuid4().hex,
                data_array=np.array([production_amount], dtype=float),
                indices_array=np.array([(node_id, node_id)], dtype=bwp.INDICES_DTYPE),
            )

        return datapackage_technosphere

    def create_biosphere_datapackage(self) -> bwp.Datapackage:
        """
        Creates the modifications to the biosphere matrix in form of a datapackage.
        Datapackages add or overwrite data points in the LCA matrices before LCA calculations.
        It adds the biosphere flows to the exploded technosphere processes.

        This function iterates over each unique producer, and for each biosphere exchange
        of the original activity, it creates a new biosphere exchange for the new "temporalized" node.

        Temporal markets have no biosphere exchanges, as they only divide the amount of
        a technosphere exchange between the different databases.

        Parameters
        ----------
        None

        Returns
        -------
        bwp.Datapackage
            A datapackage containing the modifications for the biosphere matrix.
        """
        unique_producers = (
            self.timeline.groupby(["producer", "time_mapped_producer"])
            .count()
            .index.values
        )  # array of unique ((original) producer_id, (new) time_mapped_producer_id) tuples

        datapackage_biosphere = bwp.create_datapackage(sum_inter_duplicates=False)

        for producer in unique_producers:
            original_producer_node = self.nodes[producer[0]]
            if (
                original_producer_node["database"]
                not in self.database_dates_static.keys()  # skip temporal markets
            ):
                new_producer_id = producer[1]
                indices = (
                    []
                )  # list of (biosphere, technosphere) indices for the biosphere flow exchanges
                amounts = []  # list of amounts corresponding to the bioflows
                for exc in original_producer_node.biosphere():
                    indices.append(
                        (exc.input.id, new_producer_id)
                    )  # directly build a list of tuples to pass into the datapackage, the new_producer_id is the new column index
                    amounts.append(exc.amount)

                datapackage_biosphere.add_persistent_vector(
                    matrix="biosphere_matrix",
                    name=uuid.uuid4().hex,
                    data_array=np.array(amounts, dtype=float),
                    indices_array=np.array(
                        indices,
                        dtype=bwp.INDICES_DTYPE,
                    ),
                    flip_array=np.array([False], dtype=bool),
                )
        return datapackage_biosphere

    def add_row_to_technosphere_datapackage(
        self,
        row: pd.core.frame,
        datapackage: bwp.Datapackage,
        new_nodes: set,
    ) -> None:
        """
        This adds the modifications to the technosphere matrix for each time-dependent exchange
        as datapackage elements to a given `bwp.Datapackage`.


        Modifications include:

        1) Exploded processes: new matrix elements for time-explicit consumer and time-explicit
        producer, representing the temporal edge between them.

        2) Temporal markets: new matrix entries for "temporal markets" and links to the producers
        in temporally matching background databases. Processes in the background databases are
        matched on name, reference product and location.

        3) Diagonal entries: ones on the diagonal for new nodes.

        This function also collects the ids of new nodes, temporalized nodes and temporal markets.

        Parameters
        ----------
        row : pd.core.frame
            A row of the timeline DataFrame representing an temporalized edge
        datapackage : bwp.Datapackage
            Append to this datapackage, if available. Otherwise create a new datapackage.
        new_nodes : set
            Set of tuples (node_id, production_amount) to which new node ids are added.

        Returns
        -------
        None
            Adds elements for this edge to the bwp.Datapackage and stores the ids of new nodes, temporalized nodes and temporal markets.
        """

        if row.consumer == -1:  # functional unit
            new_producer_id = row.time_mapped_producer
            fu_production_amount = self.nodes[row.producer].rp_exchange().amount
            new_nodes.add((new_producer_id, fu_production_amount))
            self.temporalized_process_ids.add(
                new_producer_id
            )  # comes from foreground, so it is a temporalized process
            return

        new_consumer_id = row.time_mapped_consumer
        new_producer_id = row.time_mapped_producer

        previous_producer_id = row.producer
        previous_producer_node = self.nodes[previous_producer_id]

        production_exchange_amount = self.nodes[row.consumer].rp_exchange().amount
        scaled_amount = row.amount * abs(
            production_exchange_amount
        )  # abs value used for scaling to preserve the sign of the exchange

        # Add entry between exploded consumer and exploded producer (not in background database)
        datapackage.add_persistent_vector(
            matrix="technosphere_matrix",
            name=uuid.uuid4().hex,
            data_array=np.array([scaled_amount], dtype=float),
            indices_array=np.array(
                [(new_producer_id, new_consumer_id)],
                dtype=bwp.INDICES_DTYPE,
            ),
            flip_array=np.array([True], dtype=bool),
        )

        # Check if previous producer comes from background database -> temporal market
        if previous_producer_node["database"] in self.database_dates_static.keys():
            # Create new edges based on temporal_market_shares from the row
            if not row.temporal_market_shares:
                raise ValueError(
                    f"Row {row} has no temporal market shares, but the previous producer {previous_producer_node['name']} is from a background database."
                )
            for database, db_share in row.temporal_market_shares.items():
                # Get the producer activity in the corresponding background database
                try:
                    producer_id_in_background_db = (
                        self.interdatabase_activity_mapping.find_match(
                            previous_producer_id, database
                        )
                    )
                except KeyError as exc:
                    raise ValueError(
                        f"Producer with ID {previous_producer_id} not found in database {database}."
                    ) from exc
                except Exception as e:
                    raise RuntimeError(
                        f"Error finding producer {previous_producer_id} in database {database}: {e}"
                    ) from e

                # Add entry between exploded producer and producer in background database
                # -->("Temporal Market")
                datapackage.add_persistent_vector(
                    matrix="technosphere_matrix",
                    name=uuid.uuid4().hex,
                    data_array=np.array(
                        [db_share], dtype=float
                    ),  # temporal markets produce 1, so shares divide amount between dbs
                    indices_array=np.array(
                        [(producer_id_in_background_db, new_producer_id)],
                        dtype=bwp.INDICES_DTYPE,
                    ),
                    flip_array=np.array([True], dtype=bool),
                )
                self.temporal_market_ids.add(new_producer_id)
                producer_production_amount = (
                    1  # Shares sum up to 1, so production amount is 1
                )

        else:  # comes from foreground, so it is a temporalized process
            self.temporalized_process_ids.add(new_producer_id)

            # Get the production amount of the previous producer if it's not a temporal market - needed for diagonal matrix entry
            if isinstance(
                previous_producer_node, bd.backends.iotable.proxies.IOTableActivity
            ):
                if len(previous_producer_node.production()) == 1:
                    producer_production_amount = list(
                        previous_producer_node.production()
                    )[0].amount
                else:
                    raise ValueError(
                        "The producer activity is of type IOTableActivity, but has more than one production exchange. This is currently not supported."
                    )
            elif isinstance(previous_producer_node, bd.backends.proxies.Activity):
                producer_production_amount = (
                    self.nodes[row.producer].rp_exchange().amount
                )
            else:
                raise ValueError(
                    f"Can't determine the production amount of the producer activity {previous_producer_node['name']} , as it's of an unknown type."
                )

        # Add newly created producing process to new_nodes
        new_nodes.add((new_producer_id, producer_production_amount))
